
__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import time
import threading
import pymongo

import radical.utils as ru

from .. import utils     as rpu
from .. import constants as rpc


# ==============================================================================
#
DEFAULT_BULK_COLLECTION_TIME =  1.0  # seconds
DEFAULT_BULK_COLLECTION_SIZE =  100  # seconds


# ==============================================================================
#
class Update(rpu.Worker):
    """
    An UpdateWorker pushes CU and Pilot state updates to mongodb.  Its instances
    compete for update requests on the update_queue.  Those requests will be
    triplets of collection name, query dict, and update dict.  Update requests
    will be collected into bulks over some time (BULK_COLLECTION_TIME) and
    number (BULK_COLLECTION_SIZE) to reduce number of roundtrips.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid = ru.generate_id('update.%(counter)s', ru.ID_CUSTOM)

        rpu.Worker.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def create(cls, cfg, session):

        return cls(cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

        self._session_id = self._cfg['session_id']
        self._dburl      = self._cfg['dburl']
        self._owner      = self._cfg['owner']

        # TODO: get db handle from a connected session
        _, db, _, _, _   = ru.mongodb_connect(self._dburl)
        self._mongo_db   = db
        self._coll       = self._mongo_db[self._session_id]
        self._bulk       = self._coll.initialize_ordered_bulk_op()
        self._last       = time.time()        # time of last bulk push
        self._uids       = list()             # list of collected uids
        self._lock       = threading.RLock()  # protect _bulk

        self._bct        = self._cfg.get('bulk_collection_time',
                                          DEFAULT_BULK_COLLECTION_TIME)
        self._bcs        = self._cfg.get('bulk_collection_size',
                                          DEFAULT_BULK_COLLECTION_SIZE)

        self.register_subscriber(rpc.STATE_PUBSUB, self._state_cb)
        self.register_timed_cb(self._idle_cb, timer=self._bct)


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._session._log.debug('%s stop called', self._uid)
        super(Update, self).stop()


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        self.unregister_timed_cb(self._idle_cb)
        self.unregister_subscriber(rpc.STATE_PUBSUB, self._state_cb)


    # --------------------------------------------------------------------------
    #
    def _timed_bulk_execute(self, flush=False):

        # is there anything to execute?
        if not self._uids:
            return False

        now = time.time()
        age = now - self._last

        # only push if flush is forced, or when collection time or size 
        # have been exceeded
        if  not flush \
            and age < self._bct \
            and len(self._uids) < self._bcs:
            return False

        try:
            res = self._bulk.execute()
            self._log.debug("bulk update result: %s", res)
        except pymongo.errors.OperationFailure as e:
            self._log.exception('bulk exec error: %s' % e.details)
            raise
        except Exception as e:
            self._log.exception('mongodb error: %s', e)
            raise

        self._prof.prof('update_pushed', msg='bulk size: %d' % len(self._uids))

        for entry in self._uids:

            uid   = entry[0]
            state = entry[2]

            if state:
                self._prof.prof('update_pushed', uid=uid, msg=state)
            else:
                self._prof.prof('update_pushed', uid=uid)

        # empty bulk, refresh state
        self._last = now
        self._bulk = self._coll.initialize_ordered_bulk_op()
        self._uids = list()

        return True


    # --------------------------------------------------------------------------
    #
    def _idle_cb(self):

        with self._lock:
             self._timed_bulk_execute()

        return True


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, topic, msg):
        """

        # FIXME: this documentation is not final, nor does it reflect reality!

        'msg' is expected to be of the form ['cmd', 'thing'], where 'thing' is
        an entity to update in the DB, and 'cmd' specifies the mode of update.

        'things' are expected to be dicts with a 'type' and 'uid' field.  If
        either one does not exist, an exception is raised.

        Supported types are:

          - unit
          - pilot

        supported 'cmds':

          - delete      : delete can be delayed until bulk is collected/flushed
          - update      : update can be delayed until bulk is collected/flushed
          - state       : update can be delayed until bulk is collected/flushed
                          only state and state history are updated
          - delete_flush: delete is sent immediately (possibly in a bulk)
          - update_flush: update is sent immediately (possibly in a bulk)
          - state_flush : update is sent immediately (possibly in a bulk)
                          only state and state history are updated
          - flush       : flush pending bulk

        The 'thing' can contains '$set' and '$push' fields, which will then be
        used as given.  For all other fields, we use the following convention:

          - scalar values: use '$set'
          - dict   values: use '$set'
          - list   values: use '$push'

        That implies that all potential 'list' types should be defined in the
        initial 'thing' insert as such, as (potentially empty) lists.

        For 'cmd' in ['state', 'state_flush'], only the 'uid' and 'state' fields
        of the given 'thing' are used, all other fields are ignored.  If 'state'
        does not exist, an exception is raised.
        """

        cmd    = msg['cmd']
        things = msg['arg']

      # cmds = ['delete',       'update',       'state',
      #         'delete_flush', 'update_flush', 'state_flush', 'flush']
        if cmd not in ['update']:
            self._log.info('ignore cmd %s', cmd)
            return True

        if not isinstance(things, list):
            things = [things]


        # FIXME: we don't have any error recovery -- any failure to update
        #        state in the DB will thus result in an exception here and tear
        #        down the module.
        for thing in things:

            # got a new request.  Add to bulk (create as needed),
            # and push bulk if time is up.
            uid   = thing['uid']
            ttype = thing['type']
            state = thing['state']

            if 'clone' in uid:
                # we don't push clone states to DB
                return True

            self._prof.prof('update_request', msg=state, uid=uid)

            if not state:
                # nothing to push
                return True

            # create an update document
            update_dict          = dict()
            update_dict['$set']  = dict()
            update_dict['$push'] = dict()

            for key,val in thing.iteritems():
                # we never set _id, states (to avoid index clash, duplicated ops)
                if key not in ['_id', 'states']:
                    update_dict['$set'][key] = val

            # we set state, put (more importantly) we push the state onto the
            # 'states' list, so that we can later get state progression in sync with
            # the state model, even if they have been pushed here out-of-order
            update_dict['$push']['states'] = state

            with self._lock:

                # push the update request onto the bulk
                self._uids.append([uid, ttype, state])
                self._bulk.find  ({'uid'  : uid, 
                                   'type' : ttype}) \
                          .update(update_dict)

        with self._lock:
            # attempt a timed update
            self._timed_bulk_execute()

        return True


# ------------------------------------------------------------------------------






__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

# ------------------------------------------------------------------------------
#
# This file contains methods to support the transition of the non-gpu aware
# scheduling data structures to a gpu aware version.  It is used to allow
# combine new AgentScheduling implementations with 'old' LaunchMethods in the
# AgentExecutingComponent. The LMs have to make sure to end up with a version of
# the slot structure they can handle.
#
#     slots = { ...
#       'cpu_processes'    : 4,
#       'cpu_process_type' : 'mpi',
#       'cpu_threads'      : 2,
#       'gpu_processes     : 2,
#       'slots' :
#       {                 # [[nodename, [node_uid], [core indexes],   [gpu idx]]]
#         'nodes'         : [[node_1,   node_uid_1, [[0, 2], [4, 6]], [[0]    ]],
#                            [node_2,   node_uid_2, [[1, 3], [5, 7]], [[0]    ]]],
#         'cores_per_node': 8,
#         'gpus_per_node' : 1,
#         'lm_info'       : { ... }
#       }
#     }
#
def slots2opaque(slots, nodes):
    '''
    Translate the new `slots` structure as documented in the agent scheduler's
    base class into the old `opaque_slots` structure as documented below.
    '''

    # old opaque_slots structure:
    #
    # task_slots = [self.nodes[x]['name'] for x in offsets]
    # opaque_slots = {
    #     'task_slots'   : task_slots,         # [node:core]
    #     'task_offsets' : offsets,            # [core] (global)
    #     'lm_info'      : self._lrms_lm_info  # unchanged
    # }
    #
    task_slots    = list()
    task_offsets  = list()

    for entry in slots:
        for cpu_set in entry[2]:
            task_slots.append('%s:%d' % (entry[1], cpu_set[0]))

        for gpu_set in entry[3]:
            task_slots.append('%s:%d' % (entry[1], gpu_set[0]))

    task_offsets = _slots2offset(task_slots, nodes)

    return {'task_slots'   : task_slots,
            'task_offsets' : task_offsets,
            'lm_info'      : slots['lrms_lm_info']}


# --------------------------------------------------------------------------
#
# Convert a set of slots into an index into the global slots list
#
def _slots2offset(task_slots, nodes):
    '''
    Back-translate the opaque_slots derived above into global core offsets
    '''

    # Note: rhis assumes all hosts have the same number of cores

    global_idx     = 0
    global_offsets = list()

    for node in nodes:

        cores_per_node = len(node['cores'])

        for ts in task_slots:

            node_name, core_idx = ts.split(':', 1)

            if node_name == node['uid']:
                global_offsets.append(global_idx + core_idx)

        global_idx += cores_per_node

    return min(global_offsets)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
# we *first* import radical.utils, so that the monkeypatching of the logger has
# a chance to kick in before the logging module is pulled by any other 3rd party
# module, and also to monkeypatch `os.fork()` for the `atfork` functionality
import radical.utils as _ru

# ------------------------------------------------------------------------------
# constants and types
from .types      import *
from .states     import *
from .logentry   import * 
from .constants  import *
from .exceptions import *


# ------------------------------------------------------------------------------
# import API
from .session                   import Session
from .context                   import Context

from .unit_manager              import UnitManager
from .compute_unit              import ComputeUnit
from .compute_unit_description  import ComputeUnitDescription
from .compute_unit_description  import POSIX, MPI, OpenMP

from .pilot_manager             import PilotManager
from .compute_pilot             import ComputePilot
from .compute_pilot_description import ComputePilotDescription

from .resource_config           import ResourceConfig
from .staging_directives        import COPY, LINK, MOVE, TRANSFER
from .staging_directives        import SKIP_FAILED, CREATE_PARENTS

from .utils                     import version, version_short
from .utils                     import version_detail, version_branch
from .utils                     import sdist_name, sdist_path


# ------------------------------------------------------------------------------
# make submodules available -- mostly for internal use
import utils
import worker
import umgr
import pmgr
import agent


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2014-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import os
import sys
import copy
import stat
import time
import pprint
import subprocess         as sp

import radical.utils      as ru

from ..  import utils     as rpu
from ..  import states    as rps
from ..  import constants as rpc
from ..  import Session   as rp_Session

from .  import rm         as rpa_rm
from .  import lm         as rpa_lm


# this needs git attribute 'ident' set for this file
git_ident = '$Id$'


# ==============================================================================
#
class Agent_0(rpu.Worker):

    # This is the base agent.  It does not do much apart from starting
    # sub-agents and watching them  If any of the sub-agents die, it will shut
    # down the other sub-agents and itself.
    #
    # This class inherits the rpu.Worker, so that it can use the communication
    # bridges and callback mechanisms.  It will own a session (which creates said
    # communication bridges (or at least some of them); and a controller, which
    # will control the sub-agents.

    # --------------------------------------------------------------------------
    #
    def __init__(self, agent_name):

        assert(agent_name == 'agent_0'), 'expect agent_0, not subagent'
        print 'startup agent %s' % agent_name

        # load config, create session, init rpu.Worker
        agent_cfg  = '%s/%s.cfg' % (os.getcwd(), agent_name)
        cfg        = ru.read_json_str(agent_cfg)

        cfg['agent_name'] = agent_name

        self._uid         = agent_name
        self._pid         = cfg['pilot_id']
        self._sid         = cfg['session_id']
        self._runtime     = cfg['runtime']
        self._starttime   = time.time()
        self._final_cause = None
        self._lrms        = None

        # this better be on a shared FS!
        cfg['workdir']    = os.getcwd()

        # sanity check on config settings
        if 'cores'               not in cfg: raise ValueError('Missing number of cores')
        if 'lrms'                not in cfg: raise ValueError('Missing LRMS')
        if 'dburl'               not in cfg: raise ValueError('Missing DBURL')
        if 'pilot_id'            not in cfg: raise ValueError('Missing pilot id')
        if 'runtime'             not in cfg: raise ValueError('Missing or zero agent runtime')
        if 'scheduler'           not in cfg: raise ValueError('Missing agent scheduler')
        if 'session_id'          not in cfg: raise ValueError('Missing session id')
        if 'spawner'             not in cfg: raise ValueError('Missing agent spawner')
        if 'task_launch_method'  not in cfg: raise ValueError('Missing unit launch method')

        # Check for the RADICAL_PILOT_DB_HOSTPORT env var, which will hold
        # the address of the tunnelized DB endpoint. If it exists, we
        # overrule the agent config with it.
        hostport = os.environ.get('RADICAL_PILOT_DB_HOSTPORT')
        if hostport:
            dburl = ru.Url(cfg['dburl'])
            dburl.host, dburl.port = hostport.split(':')
            cfg['dburl'] = str(dburl)

        # Create a session.
        #
        # This session will connect to MongoDB, and will also create any
        # communication channels and components/workers specified in the
        # config -- we merge that information into our own config.
        # We don't want the session to start components though, so remove them
        # from the config copy.        
        session_cfg = copy.deepcopy(cfg)
        session_cfg['components'] = dict()
        session = rp_Session(cfg=session_cfg, uid=self._sid)

        # we still want the bridge addresses known though, so make sure they are
        # merged into our own copy, along with any other additions done by the
        # session.
        ru.dict_merge(cfg, session._cfg, ru.PRESERVE)
        pprint.pprint(cfg)

        if not session.is_connected:
            raise RuntimeError('agent_0 could not connect to mongodb')

        # at this point the session is up and connected, and it should have
        # brought up all communication bridges and the UpdateWorker.  We are
        # ready to rumble!
        rpu.Worker.__init__(self, cfg, session)

        # this is the earlier point to sync bootstrapper and agent # profiles
        self._prof.prof('sync_rel', msg='agent_0 start', uid=self._pid)

        # Create LRMS which will give us the set of agent_nodes to use for
        # sub-agent startup.  Add the remaining LRMS information to the
        # config, for the benefit of the scheduler).
        self._lrms = rpa_rm.RM.create(name=self._cfg['lrms'], cfg=self._cfg,
                                      session=self._session)

        # add the resource manager information to our own config
        self._cfg['lrms_info'] = self._lrms.lrms_info


    # --------------------------------------------------------------------------
    #
    def initialize_parent(self):

        # create the sub-agent configs
        self._write_sa_configs()

        # and start the sub agents
        self._start_sub_agents()

        # register the command callback which pulls the DB for commands
        self.register_timed_cb(self._agent_command_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # registers the staging_input_queue as this is what we want to push
        # units to
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.AGENT_STAGING_INPUT_QUEUE)

        # sub-agents are started, components are started, bridges are up: we are
        # ready to roll!
        pilot = {'type'             : 'pilot',
                 'uid'              : self._pid,
                 'state'            : rps.PMGR_ACTIVE,
                 'resource_details' : {
                     'lm_info'      : self._lrms.lm_info.get('version_info'),
                     'lm_detail'    : self._lrms.lm_info.get('lm_detail')},
                 '$set'             : ['resource_details']}
        self.advance(pilot, publish=True, push=False)

        # register idle callback to pull for units -- which is the only action
        # we have to perform, really
        self.register_timed_cb(self._check_units_cb,
                               timer=self._cfg['db_poll_sleeptime'])


        # record hostname in profile to enable mapping of profile entries
        self._prof.prof(event='hostname', uid=self._pid, msg=ru.get_hostname())


    # --------------------------------------------------------------------------
    #
    def finalize_parent(self):

        # tear things down in reverse order
        self._prof.flush()
        self._log.info('publish "terminate" cmd')
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'terminate',
                                          'arg' : None})

        self.unregister_timed_cb(self._check_units_cb)
        self.unregister_output(rps.AGENT_STAGING_INPUT_PENDING)
        self.unregister_timed_cb(self._agent_command_cb)

        if self._lrms:
            self._log.debug('stop    lrms %s', self._lrms)
            self._lrms.stop()
            self._log.debug('stopped lrms %s', self._lrms)

        if   self._final_cause == 'timeout'  : state = rps.DONE
        elif self._final_cause == 'cancel'   : state = rps.CANCELED
        elif self._final_cause == 'sys.exit' : state = rps.CANCELED
        else                                 : state = rps.FAILED

        self._log.debug('final state: %s (%s)', state, self._final_cause)
      # # we don't rely on the existence / viability of the update worker at
      # # that point.
      # FIXME:
      # self._log.debug('update db state: %s: %s', state, self._final_cause)
      # self._update_db(state, self._final_cause)


    # --------------------------------------------------------------------------
    #
    def wait_final(self):

        while self._final_cause is None:
          # self._log.info('no final cause -> alive')
            time.sleep(1)

        self._log.debug('final: %s', self._final_cause)

      # if self._session:
      #     self._log.debug('close  session %s', self._session.uid)
      #     self._session.close()
      #     self._log.debug('closed session %s', self._session.uid)


    # --------------------------------------------------------------------------
    #
    def _update_db(self, state, msg=None):

        # NOTE: we do not push the final pilot state, as that is done by the
        #       bootstrapper *after* this poilot *actually* finished.

        self._log.info('pilot state: %s', state)
        self._log.info('rusage: %s', rpu.get_rusage())
        self._log.info(msg)

        if state == rps.FAILED:
            self._log.info(ru.get_trace())

        out = None
        err = None
        log = None

        try    : out = open('./agent_0.out', 'r').read(1024)
        except Exception: pass
        try    : err = open('./agent_0.err', 'r').read(1024)
        except Exception: pass
        try    : log = open('./agent_0.log', 'r').read(1024)
        except Exception: pass

        ret = self._session._dbs._c.update(
                {'type'   : 'pilot',
                 'uid'    : self._pid},
                {'$set'   : {'stdout'        : rpu.tail(out),
                             'stderr'        : rpu.tail(err),
                             'logfile'       : rpu.tail(log)}
                })
        self._log.debug('update ret: %s', ret)


    # --------------------------------------------------------------------------
    #
    def _write_sa_configs(self):

        # we have all information needed by the subagents -- write the
        # sub-agent config files.

        # write deep-copies of the config for each sub-agent (sans from agent_0)
        for sa in self._cfg.get('agents', {}):

            assert(sa != 'agent_0'), 'expect subagent, not agent_0'

            # use our own config sans agents/components as a basis for
            # the sub-agent config.
            tmp_cfg = copy.deepcopy(self._cfg)
            tmp_cfg['agents']     = dict()
            tmp_cfg['components'] = dict()

            # merge sub_agent layout into the config
            ru.dict_merge(tmp_cfg, self._cfg['agents'][sa], ru.OVERWRITE)

            tmp_cfg['agent_name'] = sa
            tmp_cfg['owner']      = 'agent_0'

            ru.write_json(tmp_cfg, './%s.cfg' % sa)


    # --------------------------------------------------------------------------
    #
    def _start_sub_agents(self):
        '''
        For the list of sub_agents, get a launch command and launch that
        agent instance on the respective node.  We pass it to the seconds
        bootstrap level, there is no need to pass the first one again.
        '''

        # FIXME: we need a watcher cb to watch sub-agent state

        self._log.debug('start_sub_agents')

        if not self._cfg.get('agents'):
            self._log.debug('start_sub_agents noop')
            return

        # the configs are written, and the sub-agents can be started.  To know
        # how to do that we create the agent launch method, have it creating
        # the respective command lines per agent instance, and run via
        # popen.
        #
        # actually, we only create the agent_lm once we really need it for
        # non-local sub_agents.
        agent_lm   = None
        for sa in self._cfg['agents']:

            target = self._cfg['agents'][sa]['target']

            if target == 'local':

                # start agent locally
                cmdline = '/bin/sh -l %s/bootstrap_2.sh %s' % (os.getcwd(), sa)

            elif target == 'node':

                if not agent_lm:
                    agent_lm = rpa_lm.LaunchMethod.create(
                        name    = self._cfg['agent_launch_method'],
                        cfg     = self._cfg,
                        session = self._session)

                node = self._cfg['lrms_info']['agent_nodes'][sa]
                # start agent remotely, use launch method
                # NOTE:  there is some implicit assumption that we can use
                #        the 'agent_node' string as 'agent_string:0' and
                #        obtain a well format slot...
                # FIXME: it is actually tricky to translate the agent_node
                #        into a viable 'slots' structure, as that is
                #        usually done by the schedulers.  So we leave that
                #        out for the moment, which will make this unable to
                #        work with a number of launch methods.  Can the
                #        offset computation be moved to the LRMS?
                ls_name = "%s/%s.sh" % (os.getcwd(), sa)
                slots = {
                  'cpu_processes' : 1,
                  'cpu_threads'   : 1,
                  'gpu_processes' : 0,
                  'gpu_threads'   : 0,
                # 'nodes'         : [[node[0], node[1], [[0]], []]],
                  'nodes'         : [{'name'    : node[0], 
                                     'uid'     : node[1],
                                     'core_map': [[0]],
                                     'gpu_map' : [],
                                     'lfs'     : {'path': '/tmp', 'size': 0}
                                    }],
                  'cores_per_node': self._cfg['lrms_info']['cores_per_node'],
                  'gpus_per_node' : self._cfg['lrms_info']['gpus_per_node'],
                  'lm_info'       : self._cfg['lrms_info']['lm_info']
                }
                agent_cmd = {
                        'uid'          : sa,
                        'slots'        : slots,
                        'description'  : {
                            'cpu_processes' : 1,
                            'executable'    : "/bin/sh",
                            'mpi'           : False,
                            'arguments'     : ["%s/bootstrap_2.sh" % os.getcwd(), sa]
                            }
                        }
                cmd, hop = agent_lm.construct_command(agent_cmd,
                        launch_script_hop='/usr/bin/env RP_SPAWNER_HOP=TRUE "%s"' % ls_name)

                with open (ls_name, 'w') as ls:
                    # note that 'exec' only makes sense if we don't add any
                    # commands (such as post-processing) after it.
                    ls.write('#!/bin/sh\n\n')
                    ls.write('exec %s\n' % cmd)
                    st = os.stat(ls_name)
                    os.chmod(ls_name, st.st_mode | stat.S_IEXEC)

                if hop : cmdline = hop
                else   : cmdline = ls_name

            # spawn the sub-agent
            self._log.info ('create sub-agent %s: %s' % (sa, cmdline))

            # ------------------------------------------------------------------
            class _SA(ru.Process):
                def __init__(self, sa, cmd, log):
                    self._sa   = sa
                    self._cmd  = cmd.split()
                    self._log  = log
                    self._proc = None
                    super(_SA, self).__init__(name=sa, log=self._log)
                    self.start()

                def ru_initialize_child(self):
                    sys.stdout = open('%s.out' % self._ru_name, 'w')
                    sys.stderr = open('%s.err' % self._ru_name, 'w')
                    out = open('%s.out' % self._sa, 'w')
                    err = open('%s.err' % self._sa, 'w')
                    self._proc = sp.Popen(args=self._cmd, stdout=out, stderr=err)

                def work_cb(self):
                    time.sleep(0.1)
                    if self._proc.poll() is None:
                        return True   # all is well
                    else:
                        return False  # proc is gone - terminate

                def ru_finalize_child(self):
                    if self._proc:
                        try:
                            self._proc.terminate()
                        except Exception as e:
                            # we are likely racing on termination...
                            self._log.warn('%s term failed: %s', self._sa, e)
            # ------------------------------------------------------------------

            # the agent is up - let the watcher manage it from here
            self.register_watchable(_SA(sa, cmdline, log=self._log))

        self._log.debug('start_sub_agents done')


    # --------------------------------------------------------------------------
    #
    def _agent_command_cb(self):

        self.is_valid()

        if not self._check_commands(): return False
        if not self._check_state   (): return False

        return True


    # --------------------------------------------------------------------------
    #
    def _check_commands(self):

        # Check if there's a command waiting
        # FIXME: this pull should be done by the update worker, and commands
        #        should then be communicated over the command pubsub
        # FIXME: commands go to pmgr, umgr, session docs
        # FIXME: this is disabled right now
        retdoc = self._session._dbs._c.find_and_modify(
                    query ={'uid'  : self._pid},
                    update={'$set' : {'cmd': []}},  # Wipe content of array
                    fields=['cmd'])

        if not retdoc:
            return True  # this is not an error

        for spec in retdoc.get('cmd', []):

            cmd = spec['cmd']
            arg = spec['arg']

            self._prof.prof('cmd', msg="%s : %s" % (cmd, arg), uid=self._pid)

            if cmd == 'heartbeat':
                self._log.info('heartbeat_in')


            elif cmd == 'cancel_pilot':
                self._log.info('cancel pilot cmd')
                self._log.info('publish "terminate" cmd')
                self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'terminate',
                                                  'arg' : None})

              # self.stop()
                self._ru_term.set()

                with open('./killme.signal', 'w+') as f:
                    f.write(rps.CANCELED)
                    f.flush()

                self._final_cause = 'cancel'
                return False  # we are done

            elif cmd == 'cancel_units':

                self._log.info('cancel_units cmd')
                self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'cancel_units',
                                                  'arg' : arg})
            else:
                self._log.error('could not interpret cmd "%s" - ignore', cmd)

        return True


    # --------------------------------------------------------------------------
    #
    def _check_state(self):

        # Make sure that we haven't exceeded the runtime (if one is set). If
        # we have, terminate.
        if self._runtime:
            if time.time() >= self._starttime + (int(self._runtime) * 60):
                self._log.info('reached runtime limit (%ss).', self._runtime*60)
                self._final_cause = 'timeout'
                self.stop()
                return False  # we are done

        return True


    # --------------------------------------------------------------------------
    #
    def _check_units_cb(self):

        self.is_valid()

        # FIXME: this should probably go into a custom `is_valid()`
        if not self._session._dbs._c:
            self._log.warn('db connection gone - abort')
            return False

        # Check if there are compute units waiting for input staging
        # and log that we pulled it.
        #
        # FIXME: Unfortunately, 'find_and_modify' is not bulkable, so we have
        #        to use 'find'.  To avoid finding the same units over and over
        #        again, we update the 'control' field *before* running the next
        #        find -- so we do it right here.
        #        This also blocks us from using multiple ingest threads, or from
        #        doing late binding by unit pull :/
        unit_cursor = self._session._dbs._c.find({'type'    : 'unit',
                                                  'pilot'   : self._pid,
                                                  'control' : 'agent_pending'})
        if not unit_cursor.count():
            # no units whatsoever...
            self._log.info('units pulled:    0')
            return True  # this is not an error

        # update the units to avoid pulling them again next time.
        unit_list = list(unit_cursor)
        unit_uids = [unit['uid'] for unit in unit_list]

        self._log.info('units PULLED: %4d', len(unit_list))

        self._session._dbs._c.update({'type'  : 'unit',
                                      'uid'   : {'$in'     : unit_uids}},
                                     {'$set'  : {'control' : 'agent'}},
                                     multi=True)

        self._log.info("units pulled: %4d", len(unit_list))
        self._prof.prof('get', msg='bulk size: %d' % len(unit_list),
                        uid=self._pid)

        for unit in unit_list:

            # we need to make sure to have the correct state:
            unit['state'] = rps._unit_state_collapse(unit['states'])
            self._prof.prof('get', uid=unit['uid'])

            # FIXME: raise or fail unit!
            if unit['control'] != 'agent_pending':
                self._log.error('invalid control: %s', (pprint.pformat(unit)))

            if unit['state'] != rps.AGENT_STAGING_INPUT_PENDING:
                self._log.error('invalid state: %s', (pprint.pformat(unit)))

            unit['control'] = 'agent'

        # now we really own the CUs, and can start working on them (ie. push
        # them into the pipeline).  We don't publish nor profile as advance,
        # since that happened already on the module side when the state was set.
        self.advance(unit_list, publish=False, push=True)

        return True


# ------------------------------------------------------------------------------




import os
import sys



number_of_folders = int(sys.argv[1])

for i in xrange(number_of_folders):
    directory = '/tmp/stage_in_folder_%d' % i 
    if not os.path.exists(directory):
        os.makedirs(directory)

    filename = '/tmp/stage_in_folder_%d/input_file.dat'%i
    afile = open(filename,'w')
    afile.write('hello world!')
    afile.close()




__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from .base import LaunchMethod



# ==============================================================================
#
# dplace: job launcher for SGI systems (e.g. on Blacklight)
# https://www.nas.nasa.gov/hecc/support/kb/Using-SGIs-dplace-Tool-for-Pinning_284.html
#
class MPIRunDPlace(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        LaunchMethod.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self.launch_command = ru.which([
            'mpirun',            # General case
            'mpirun_rsh',        # Gordon @ SDSC
            'mpirun-mpich-mp',   # Mac OSX MacPorts
            'mpirun-openmpi-mp'  # Mac OSX MacPorts
        ])

        self.dplace_command = ru.which([
            'dplace',            # General case
        ])

        if not self.dplace_command:
            raise RuntimeError("mpirun not found!")

        self.mpi_version, self.mpi_flavor = self._get_mpi_info(self.launch_command)


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_threads = cud['cpu_thread']
        task_env     = cud.get('environment') or dict()
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        if task_threads > 1:
            # dplace pinning would disallow threads to map to other cores
            raise ValueError('dplace can not place threads [%d]' % task_threads)

        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
        env_list   = self.EXPORT_ENV_VARIABLES + task_env.keys()
        if env_list:

            if self.mpi_flavor == self.MPI_FLAVOR_HYDRA:
                env_string = '-envlist "%s"' % ','.join(env_list)

            elif self.mpi_flavor == self.MPI_FLAVOR_OMPI:
                for var in env_list:
                    env_string += '-x "%s" ' % var

        if 'nodes' not in slots:
            raise RuntimeError('insufficient information to launch via %s: %s'
                              % (self.name, slots))

        host_list = list()
        core_list = list()
        for node in slots['nodes']:
            tmp_list = list()
            for cpu_proc in node['core_map']:
                tmp_list.append(cpu_proc[0])
                host_list.append(node['name'])
            for gpu_proc in node['gpu_map']:
                tmp_list.append(gpu_proc[0])
                host_list.append(node['name'])

            if core_list:
                if sorted(core_list) != sorted(tmp_list):
                    raise ValueError('dplace expects heterogeneous layouts')
            else:
                core_list = tmp_list

        np          = len(host_list) + len(core_list)
        host_string = ",".join(host_list)  # FIXME: differs for mpich/hydra
        core_string = ','.join(core_list)


        command = "%s -h %s -np %d %s -c %s %s %s" % \
                  (self.launch_command, host_string, np,
                   self.dplace_command, core_string,
                   env_string, task_command)

        return command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"

import math
import multiprocessing

from .base import LRMS


# ==============================================================================
#
class Fork(LRMS):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        LRMS.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self._log.info("Using fork on localhost.")

        # For the fork LRMS (ie. on localhost), we fake an infinite number of
        # cores, so don't perform any sanity checks.
        detected_cpus = multiprocessing.cpu_count()

        if detected_cpus != self.requested_cores:
            self._log.info("using %d instead of physically available %d cores.",
                    self.requested_cores, detected_cpus)

        # if cores_per_node is set in the agent config, we slice the number of
        # cores into that many virtual nodes.  cpn defaults to requested_cores,
        # to preserve the previous behavior (1 node).
        self.cores_per_node = self._cfg.get('cores_per_node', self.requested_cores)
        self.gpus_per_node  = self._cfg.get('gpus_per_node', 0)  # FIXME GPU
        self.lfs_per_node   = {'path' : self._cfg.get('lfs_path_per_node', None),
                               'size' : self._cfg.get('lfs_size_per_node', 0)
                              }

        if not self.cores_per_node:
            self.cores_per_node = 1

        self.node_list  = list()
        requested_nodes = int(math.ceil(float(self.requested_cores) /
                                        float(self.cores_per_node ) ) )
        for i in range(requested_nodes):
            # enumerate the node list entries for a unique uis
            self.node_list.append(["localhost", 'localhost_%d' % i])

        self._log.debug('configure localhost as %s nodes (%s cores, %s gpus, %s lfs).',
                len(self.node_list), self.cores_per_node, self.gpus_per_node, self.lfs_per_node)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__ = "MIT"


import os

import radical.utils as ru

from ... import constants as rpc
from .base import AgentSchedulingComponent

import inspect
import threading as mt
from math import ceil
import logging
import pprint
# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
#
# This is an extension of the continuous scheduler with awareness of the
# file-storage capabilities on a node. The continuous data aware scheduler will
# use two data fields: availability and requirement.

# General idea:
# The availability will be obtained from the lrms_node_list and assigned to
# the node list of the class. The requirement will be obtained from the cud in
# the alloc_nompi and alloc_mpi methods. Using the availability and
# requirement, the _find_resources method will return the core and gpu ids.
#
# Expected DS of the nodelist
# self.nodes = [{
#                   'name'    : 'node_1',
#                   'uid'     : xxxx,
#                   'cores'   : 16,
#                   'gpus'    : 2,
#                   'lfs'     : 128
#               },
#               {
#                   'name'    : 'node_2',
#                   'uid'     : yyyy,
#                   'cores'   : 16,
#                   'gpus'    : 2,
#                   'lfs'     : 256
#                },
#               ]
# Q: How should the nodes be selected for MPI based units?
# lfs : in mb

import cProfile
cprof = cProfile.Profile()


def cprof_it(func):
    def wrapper(*args, **kwargs):
        retval = cprof.runcall(func, *args, **kwargs)
        return retval
    return wrapper


def dec_all_methods(dec):
    def dectheclass(cls):
        if ru.is_main_thread():
            cprof_env = os.getenv("RADICAL_PILOT_CPROFILE_COMPONENTS", "")
            cprof_elems = cprof_env.split()
            if "CONTINUOUS" in cprof_elems:
                for name, m in inspect.getmembers(cls, inspect.ismethod):
                    setattr(cls, name, dec(m))
        return cls
    return dectheclass


# ------------------------------------------------------------------------------
#
@dec_all_methods(cprof_it)
class Continuous(AgentSchedulingComponent):
    '''
    The Continuous scheduler attempts to place threads and processes of
    a compute units onto consecutive cores, gpus and nodes in the cluster.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.nodes = None
        self._tag_history = dict()

        AgentSchedulingComponent.__init__(self, cfg, session)

    # --------------------------------------------------------------------------
    #
    # FIXME: this should not be overloaded here, but in the base class
    #
    def finalize_child(self):

        cprof_env = os.getenv("RADICAL_PILOT_CPROFILE_COMPONENTS", "")
        if "CONTINUOUS" in cprof_env.split():
            self_thread = mt.current_thread()
            cprof.dump_stats("python-%s.profile" % self_thread.name)

        # make sure that parent finalizers are called
        super(Continuous, self).finalize_child()

    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # * oversubscribe:
        #   Cray's aprun for example does not allow us to oversubscribe CPU
        #   cores on a node, so we can't, say, run n CPU processes on an n-core
        #   node, and than add one additional process for a GPU application.
        # If oversubscribe` is set to False (which is the default for now),
        #   we'll prevent that behavior by allocating one additional CPU core
        #   for each set of requested GPU processes.
        #   FIXME: I think our scheme finds the wrong core IDs for GPU process
        #          startup - i.e. not the reserved ones.
        self._oversubscribe = self._cfg.get('oversubscribe', True)

        # * scattered:
        #   This is the continuous scheduler, because it attempts to allocate
        #   a *continuous* set of cores/nodes for a unit.  It does, hoewver,
        #   also allow to scatter the allocation over discontinuous nodes if
        #   this option is set.  This implementation is not optimized for the
        #   scattered mode!  The default is 'False'.
        #
        self._scattered     = self._cfg.get('scattered',     False)

        # NOTE:  for non-oversubscribing mode, we reserve a number of cores
        #        for the GPU processes - even if those GPUs are not used by
        #        a specific workload.  In this case we rewrite the node list and
        #        substract the respective number of available cores per node.
        if not self._oversubscribe:

            if self._lrms_cores_per_node <= self._lrms_gpus_per_node:
                raise RuntimeError('oversubscription mode requires more cores')

            self._lrms_cores_per_node -= self._lrms_gpus_per_node

            # since we just changed this fundamental setting, we need to
            # recreate the nodelist.
            self.nodes = []
            for node, node_uid in self._lrms_node_list:
                self.nodes.append({
                    'name': node,
                    'uid': node_uid,
                    'cores': [rpc.FREE] * self._lrms_cores_per_node,
                    'gpus': [rpc.FREE] * self._lrms_gpus_per_node,
                    'lfs': self._lrms_lfs_per_node
                })


    def _try_allocation(self, unit):
        """
        attempt to allocate cores/gpus for a specific unit.
        """

        # needs to be locked as we try to acquire slots here, but slots are
        # freed in a different thread.  But we keep the lock duration short...
        with self._slot_lock:

            self._prof.prof('schedule_try', uid=unit['uid'])
            unit['slots'] = self._allocate_slot(unit['description'])

        if unit['slots']:
            
            unit_uid = unit['uid']
            node_uids = []
            for node in unit['slots']['nodes']:
                node_uids.append(node['uid'])

            self._tag_history[unit_uid] = node_uids
            
        # the lock is freed here
        if not unit['slots']:

            # signal the unit remains unhandled (Fales signals that failure)
            self._prof.prof('schedule_fail', uid=unit['uid'])
            return False

        # got an allocation, we can go off and launch the process
        self._prof.prof('schedule_ok', uid=unit['uid'])

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug("after  allocate   %s: %s", unit['uid'],
                            self.slot_status())
            self._log.debug("%s [%s/%s] : %s", unit['uid'],
                            unit['description']['cpu_processes'],
                            unit['description']['gpu_processes'],
                            pprint.pformat(unit['slots']))

        # True signals success
        return True

    # --------------------------------------------------------------------------
    #
    def _allocate_slot(self, cud):
        '''
        This is the main method of this implementation, and is triggered when
        a unit needs to be mapped to a set of cores / gpus.  We make
        a distinction between MPI and non-MPI units (non-MPI processes MUST be
        on the same node).
        '''

        # single_node allocation is enforced for non-message passing tasks
        if cud['cpu_process_type'] == 'MPI' or \
                cud['gpu_process_type'] == 'MPI':
            slots = self._alloc_mpi(cud)
        else:
            slots = self._alloc_nompi(cud)

        if slots:
            # the unit was placed, we need to reflect the allocation in the
            # nodelist state (BUSY)
            self._change_slot_states(slots, rpc.BUSY)

        return slots

    # --------------------------------------------------------------------------
    #
    def _release_slot(self, slots):
        '''
        This method is called when previously aquired resources are not needed
        anymore.  `slots` are the resource slots as previously returned by
        `_allocate_slots()`.
        '''

        # reflect the request in the nodelist state (set to `FREE`)
        self._change_slot_states(slots, rpc.FREE)

    # --------------------------------------------------------------------------
    #
    def _find_resources(self, node, requested_cores, requested_gpus,
                        requested_lfs, core_chunk=1, partial=False,
                        lfs_chunk=1, gpu_chunk=1):
        '''
        Find up to the requested number of free cores and gpus in the node.
        This call will return two lists, for each matched set.  If the core
        does not have sufficient free resources to fulfill *both* requests, two
        empty lists are returned.  The call will *not* change the allocation
        status of the node, atomicity must be guaranteed by the caller.

        We don't care about continuity within a single node - cores `[1,5]` are
        assumed to be as close together as cores `[1,2]`.

        When `chunk` is set, only sets of exactly that size (or multiples
        thereof) are considered valid allocations.  The use case is OpenMP
        support, where each process is expected to create a certain number of
        threads, which thus must have slots on the same node available.

        NOTE: chunking is only applied to cores at this point.

        When `partial` is set to `True`, this method is allowed to return
        a *partial* match, so to find less cores, gpus, and local_fs then
        requested (but the call will never return more than requested).
        '''

        # list of core and gpu ids available in this node.
        cores = list()
        gpus = list()
        lfs = 0

        # first count the number of free cores, gpus, and local file storage.
        # This is way quicker than actually finding the core IDs.
        free_cores = node['cores'].count(rpc.FREE)
        free_gpus = node['gpus'].count(rpc.FREE)
        free_lfs = node['lfs']['size']

        alloc_lfs = alloc_cores = alloc_gpus = 0

        if partial:
            # For partial requests the check simplifies: we just check if we
            # have either, some cores *or* gpus *or* local_fs, to serve the
            # request
            if (requested_cores and not free_cores) and \
                (requested_gpus and not free_gpus) and \
                    (requested_lfs and not free_lfs):
                return [], [], None

            if requested_lfs and \
                ((requested_cores and not free_cores) and
                 (requested_gpus and not free_gpus)):
                return [], [], None

        else:
            # For non-partial requests (ie. full requests): its a no-match if
            # either the cpu or gpu request cannot be served.
            if requested_cores > free_cores or \
                requested_gpus > free_gpus or \
                    requested_lfs > free_lfs:
                return [], [], None

        # We can serve the partial or full request - alloc the chunks we need
        # FIXME: chunk gpus, too?
        # We need to land enough procs on a node such that the cores,
        # lfs and gpus requested per proc is available on the same node

        num_procs = list()

        if requested_lfs:
            alloc_lfs = min(requested_lfs, free_lfs)
            num_procs.append(alloc_lfs / lfs_chunk)
        if requested_cores:
            alloc_cores = min(requested_cores, free_cores)
            num_procs.append(alloc_cores / core_chunk)
        if requested_gpus:
            alloc_gpus = min(requested_gpus, free_gpus)
            num_procs.append(alloc_gpus / gpu_chunk)

        # Find min number of procs determined across lfs, cores, gpus
        num_procs = min(num_procs)

        # Find normalized lfs, cores and gpus
        if requested_lfs:
            alloc_lfs = num_procs * lfs_chunk
        if requested_cores:
            alloc_cores = num_procs * core_chunk
        if requested_gpus:
            alloc_gpus = num_procs * gpu_chunk

        # now dig out the core and gpu IDs.
        for idx, state in enumerate(node['cores']):

            # break if we have enough cores, else continue to pick FREE ones
            if alloc_cores == len(cores):
                break
            if state == rpc.FREE:
                cores.append(idx)

        for idx, state in enumerate(node['gpus']):

            # break if we have enough gpus, else continue to pick FREE ones
            if alloc_gpus == len(gpus):
                break
            if state == rpc.FREE:
                gpus.append(idx)

        return cores, gpus, alloc_lfs

    # --------------------------------------------------------------------------
    #
    def _get_node_maps(self, cores, gpus, threads_per_proc):
        """
        For a given set of cores and gpus, chunk them into sub-sets so that
        each sub-set can host one application process and all threads of that
        process.  Note that we currently consider all GPU applications to be
        single-threaded.
        For more details, see top level comment of `base.py`.
        """

        core_map = list()
        gpu_map = list()

        # make sure the core sets can host the requested number of threads
        assert(not len(cores) % threads_per_proc)
        n_procs = len(cores) / threads_per_proc

        idx = 0
        for p in range(n_procs):
            p_map = list()
            for t in range(threads_per_proc):
                p_map.append(cores[idx])
                idx += 1
            core_map.append(p_map)

        if idx != len(cores):
            self._log.debug('%s -- %s -- %s -- %s',
                            idx, len(cores), cores, n_procs)
        assert(idx == len(cores))

        # gpu procs are considered single threaded right now (FIXME)
        for g in gpus:
            gpu_map.append([g])

        return core_map, gpu_map

    # --------------------------------------------------------------------------
    #
    def _alloc_nompi(self, cud):
        """
        Find a suitable set of cores and gpus *within a single node*.

        Input:
        cud: Compute Unit description. Needs to specify at least one CPU
        process and one thread per CPU process, or one GPU process.
        """

        # dig out the allocation request details
        requested_procs = cud['cpu_processes']
        threads_per_proc = cud['cpu_threads']
        requested_gpus = cud['gpu_processes']
        requested_lfs = cud['lfs_per_process']
        lfs_chunk = requested_lfs if requested_lfs > 0 else 1
        tag = cud.get('tag')
        uid = cud.get('uid')

        # make sure that processes are at least single-threaded
        if not threads_per_proc:
            threads_per_proc = 1

        # cores needed for all threads and processes
        requested_cores = requested_procs * threads_per_proc

        # make sure that the requested allocation fits on a single node
        if requested_cores > self._lrms_cores_per_node or \
                requested_gpus > self._lrms_gpus_per_node or \
                requested_lfs > self._lrms_lfs_per_node['size']:

            txt = 'Non-mpi unit does not fit onto single node. \n'
            txt += 'requested cores=%s; available cores=%s \n' % (
                requested_cores, self._lrms_cores_per_node)
            txt += 'requested gpus=%s; available gpus=%s \n' % (
                requested_gpus, self._lrms_gpus_per_node)
            txt += 'requested lfs=%s; available lfs=%s' % (
                requested_lfs, self._lrms_lfs_per_node['size'])

            raise ValueError(txt)

        # ok, we can go ahead and try to find a matching node
        cores = list()
        gpus = list()
        lfs = None
        node_name = None
        node_uid = None

        for node in self.nodes:  # FIXME optimization: iteration start

            # If unit has a tag, check if the tag is in the tag_history dict,
            # else it is a invalid tag, continue as if the unit does not have
            # a tag
            # If the unit has a valid tag, find the node that matches the
            # tag from tag_history dict
            if tag and (tag in self._tag_history.keys()):
                if node['uid'] not in self._tag_history[tag]:
                    continue

            # attempt to find the required number of cores and gpus on this
            # node - do not allow partial matches.
            cores, gpus, lfs = self._find_resources(node=node,
                                                    requested_cores=requested_cores,
                                                    requested_gpus=requested_gpus,
                                                    requested_lfs=requested_lfs,
                                                    partial=False,
                                                    lfs_chunk=lfs_chunk,
                                                    core_chunk = threads_per_proc
                                                    )
            if len(cores) == requested_cores and \
                    len(gpus) == requested_gpus:

                # we found the needed resources - break out of search loop
                node_uid = node['uid']
                node_name = node['name']
                break

        # If we did not find any node to host this request, return `None`
        if not cores and not gpus and not lfs:
            return None        

        # We have to communicate to the launcher where exactly processes are to
        # be placed, and what cores are reserved for application threads.  See
        # the top level comment of `base.py` for details on the data structure
        # used to specify process and thread to core mapping.
        core_map, gpu_map = self._get_node_maps(cores, gpus, threads_per_proc)

        # We need to specify the node lfs path that the unit needs to use.
        # We set it as an environment variable that gets loaded with cud
        # executable.
        # Assumption enforced: The LFS path is the same across all nodes.
        cud['environment']['NODE_LFS_PATH'] = self._lrms_lfs_per_node['path']

        # all the information for placing the unit is acquired - return them
        slots = {'nodes': [{'name': node_name,
                            'uid': node_uid,
                            'core_map': core_map,
                            'gpu_map': gpu_map,
                            'lfs': {'size': lfs, 'path': self._lrms_lfs_per_node['path']}}],
                 'cores_per_node': self._lrms_cores_per_node,
                 'gpus_per_node': self._lrms_gpus_per_node,
                 'lfs_per_node': self._lrms_lfs_per_node,
                 'lm_info': self._lrms_lm_info
                 }

        return slots

    # --------------------------------------------------------------------------
    #
    #
    def _alloc_mpi(self, cud):
        """
        Find an available set of slots, potentially across node boundaries.  By
        default, we only allow for partial allocations on the first and last
        node - but all intermediate nodes MUST be completely used (this is the
        'CONTINUOUS' scheduler after all).

        If the scheduler is configured with `scattered=True`, then that
        constraint is relaxed, and any set of slots (be it continuous across
        nodes or not) is accepted as valid allocation.

        No matter the mode, we always make sure that we allocate in chunks of
        'threads_per_proc', as otherwise the application would not be able to
        spawn the requested number of threads on the respective node.
        """

        # dig out the allocation request details
        requested_procs = cud['cpu_processes']
        threads_per_proc = cud['cpu_threads']
        requested_gpus = cud['gpu_processes']
        requested_lfs_per_process = cud['lfs_per_process']
        tag = cud.get('tag')
        uid = cud.get('uid')

        # make sure that processes are at least single-threaded
        if not threads_per_proc:
            threads_per_proc = 1

        # cores needed for all threads and processes
        requested_cores = requested_procs * threads_per_proc

        # We allocate the same lfs per process (agreement)
        requested_lfs = requested_lfs_per_process * requested_procs

        # First and last nodes can be a partial allocation - all other nodes
        # can only be partial when `scattered` is set.
        #
        # Iterate over all nodes until we find something. Check if it fits the
        # allocation mode and sequence.  If not, start over with the next node.
        # If it matches, add the slots found and continue to next node.
        #
        # FIXME: persistent node index
        #
        # Things are complicated by chunking: we only accept chunks of
        # 'threads_per_proc', as otherwise threads would need to be distributed
        # over nodes, which is not possible for the multi-system-image clusters
        # this scheduler assumes.
        #
        #   - requested_cores > cores_per_node
        #   - cores_per_node  % threads_per_proc != 0
        #   - scattered is False
        #
        # but it can fail for less cores, too, if the partial first and last
        # allocation are not favorable.  We thus raise an exception for
        # requested_cores > cores_per_node on impossible full-node-chunking

        cores_per_node = self._lrms_cores_per_node
        gpus_per_node = self._lrms_gpus_per_node
        lfs_per_node = self._lrms_lfs_per_node

        if requested_cores > cores_per_node and \
                cores_per_node % threads_per_proc and \
                self._scattered is False:
            raise ValueError('cannot allocate under given constrains')

        # we always fail when too many threads are requested
        if threads_per_proc > cores_per_node:
            raise ValueError('too many threads requested')

        if requested_lfs_per_process > lfs_per_node['size']:
            raise ValueError('Not enough LFS for the MPI-process')

        # set conditions to find the first matching node
        is_first = True
        is_last = False
        alloced_cores = 0
        alloced_gpus = 0
        alloced_lfs = 0

        slots = {'nodes': list(),
                 'cores_per_node': cores_per_node,
                 'gpus_per_node': gpus_per_node,
                 'lfs_per_node': lfs_per_node,
                 'lm_info': self._lrms_lm_info,
                 }

        # start the search
        for node in self.nodes:

            node_uid = node['uid']
            node_name = node['name']

            # If unit has a tag, check if the tag is in the tag_history dict,
            # else it is a invalid tag, continue as if the unit does not have
            # a tag
            # If the unit has a valid tag, find the node that matches the
            # tag from tag_history dict
            if tag and (tag in self._tag_history.keys()):
                if node['uid'] not in self._tag_history[tag]:
                    continue

            # if only a small set of cores/gpus remains unallocated (ie. less
            # than node size), we are in fact looking for the last node.  Note
            # that this can also be the first node, for small units.
            if requested_cores - alloced_cores <= cores_per_node and \
                    requested_gpus - alloced_gpus <= gpus_per_node and \
                    requested_lfs - alloced_lfs <= lfs_per_node['size']:
                is_last = True

            # we allow partial nodes on the first and last node, and on any
            # node if a 'scattered' allocation is requested.
            if is_first or self._scattered or is_last:
                partial = True
            else:
                partial = False

            # now we know how many cores/gpus we still need at this point - but
            # we only search up to node-size on this node.  Duh!
            find_cores = min(requested_cores - alloced_cores, cores_per_node)
            find_gpus = min(requested_gpus - alloced_gpus,  gpus_per_node)
            find_lfs = min(requested_lfs - alloced_lfs, lfs_per_node['size'])

            # under the constraints so derived, check what we find on this node
            cores, gpus, lfs = self._find_resources(node=node,
                                                    requested_cores=find_cores,
                                                    requested_gpus=find_gpus,
                                                    requested_lfs=find_lfs,
                                                    core_chunk=threads_per_proc,
                                                    partial=partial,
                                                    lfs_chunk=requested_lfs_per_process)

            # Skip nodes that provide only lfs and no cores
            if not cores and lfs:
                continue

            if not cores and not gpus and not lfs:

                # this was not a match. If we are in  'scattered' mode, we just
                # ignore this node.  Otherwise we have to restart the search.
                if not self._scattered:
                    is_first = True
                    is_last = False
                    alloced_cores = 0
                    alloced_gpus = 0
                    alloced_lfs = 0
                    slots['nodes'] = list()

                # try next node
                continue          

            # we found something - add to the existing allocation, switch gears
            # (not first anymore), and try to find more if needed
            self._log.debug('found %s cores, %s gpus and %s lfs', cores, gpus,
                            lfs)
            core_map, gpu_map = self._get_node_maps(cores, gpus,
                                                    threads_per_proc)

            # We need to specify the node lfs path that the unit needs to use.
            # We set it as an environment variable that gets loaded with cud
            # executable.
            # Assumption enforced: The LFS path is the same across all nodes.
            if 'NODE_LFS_PATH' not in cud['environment'].keys():
                cud['environment']['NODE_LFS_PATH'] = self._lrms_lfs_per_node['path']

            slots['nodes'].append({'name': node_name,
                                   'uid': node_uid,
                                   'core_map': core_map,
                                   'gpu_map': gpu_map,
                                   'lfs': {'size': lfs, 'path': self._lrms_lfs_per_node['path']}})

            alloced_cores += len(cores)
            alloced_gpus += len(gpus)
            alloced_lfs += lfs
            is_first = False

            # or maybe don't continue the search if we have in fact enough!
            if alloced_cores == requested_cores and \
                    alloced_gpus == requested_gpus and \
                    alloced_lfs == requested_lfs:
                # we are done
                break

        # if we did not find enough, there is not much we can do at this point
        if alloced_cores < requested_cores or \
                alloced_gpus < requested_gpus or \
                alloced_lfs < requested_lfs:
            return None  # signal failure

        # this should be nicely filled out now - return
        return slots


# ------------------------------------------------------------------------------




__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

# the session needs to get rid of child process handles after forks, as Python's
# multiprocessing module does not allow to check for child process health from
# processes which did not originally spawn the children.  For htis we use
# `at_fork`, which monkeypatches `os.fork()` to support prepare, parent and
# child hooks.  We then register a child hook during session initialization.
#
# Since the monkeypatch needs to be applies before `os` is imported, we do that
# right here, in the (probably vain) hope that os was not imported before.  If
# it was, at_fork will raise an error.
#


import os
import sys
import copy
import time
import glob
import copy
import pprint
import threading

import radical.utils        as ru
import saga                 as rs
import saga.utils.pty_shell as rsup

from . import utils         as rpu
from . import states        as rps
from . import constants     as rpc
from . import types         as rpt

from .unit_manager    import UnitManager
from .pilot_manager   import PilotManager
from .resource_config import ResourceConfig
from .db              import DBSession

from .utils import version_detail as rp_version_detail


# ------------------------------------------------------------------------------
#
class Session(rs.Session):
    """
    A Session encapsulates a RADICAL-Pilot instance and is the *root* object

    A Session holds :class:`radical.pilot.PilotManager` and
    :class:`radical.pilot.UnitManager` instances which in turn hold
    :class:`radical.pilot.ComputePilot` and :class:`radical.pilot.ComputeUnit`
    instances.
    """

    # the reporter is an applicataion-level singleton
    _reporter = None

    # We keep a static typemap for component startup. If we ever want to
    # become reeeealy fancy, we can derive that typemap from rp module
    # inspection.
    #
    # --------------------------------------------------------------------------
    #
    def __init__(self, dburl=None, uid=None, cfg=None, _connect=True):
        """
        Creates a new session.  A new Session instance is created and 
        stored in the database.

        **Arguments:**
            * **dburl** (`string`): The MongoDB URL.  If none is given,
              RP uses the environment variable RADICAL_PILOT_DBURL.  If that is
              not set, an error will be raises.

            * **uid** (`string`): Create a session with this UID.  
              *Only use this when you know what you are doing!*

        **Returns:**
            * A new Session instance.

        **Raises:**
            * :class:`radical.pilot.DatabaseError`

        """

        if os.uname()[0] == 'Darwin':
            # on MacOS, we are running out of file descriptors soon.  The code
            # below attempts to increase the limit of open files - but any error
            # is silently ignored, so this is an best-effort, no guarantee.  We
            # leave responsibility for system limits with the user.
            try:
                import resource
                limits    = list(resource.getrlimit(resource.RLIMIT_NOFILE))
                limits[0] = 512
                resource.setrlimit(resource.RLIMIT_NOFILE, limits)
            except:
                pass

        self._dh          = ru.DebugHelper()
        self._valid       = True
        self._closed      = False
        self._valid_iter  = 0  # detect recursive calls of `is_valid()`

        # class state
        self._dbs         = None
        self._uid         = None
        self._dburl       = None
        self._reconnected = False

        self._cache       = dict()  # cache sandboxes etc.
        self._cache_lock  = threading.RLock()

        self._cache['resource_sandbox'] = dict()
        self._cache['session_sandbox']  = dict()
        self._cache['pilot_sandbox']    = dict()

        # before doing anything else, set up the debug helper for the lifetime
        # of the session.
        self._debug_helper = ru.DebugHelper()

        # Dictionaries holding all manager objects created during the session.
        # NOTE: should this also include agents?
        self._pmgrs      = dict()
        self._umgrs      = dict()
        self._bridges    = list()
        self._components = list()

        # FIXME: we work around some garbage collection issues we don't yet
        #        understand: instead of relying on the GC to eventually collect
        #        some stuff, we actively free those on `session.close()`, at
        #        least for the current process.  Usually, all resources get
        #        nicely collected on process termination - but not when we
        #        create many sessions (one after the other) in the same
        #        application instance (ie. the same process).  This workarounf
        #        takes care of that use case.
        #        The clean solution would be to ensure clean termination
        #        sequence, something which I seem to be unable to implement...
        #        :/
        self._to_close   = list()
        self._to_stop    = list()
        self._to_destroy = list()

        # cache the client sandbox
        # FIXME: this needs to be overwritten if configured differently in the
        #        session config, as should be the case for any agent side
        #        session instance.
        self._client_sandbox = os.getcwd()

        # The resource configuration dictionary associated with the session.
        self._resource_configs = {}

        # if a config is given, us its values:
        if cfg:
            self._cfg = copy.deepcopy(cfg)
        else:
            # otherwise we need a config
            self._cfg = ru.read_json("%s/configs/session_%s.json" \
                    % (os.path.dirname(__file__),
                       os.environ.get('RADICAL_PILOT_SESSION_CFG', 'default')))

        # fall back to config data where possible
        # sanity check on parameters
        if not uid : 
            uid = self._cfg.get('session_id')

        if uid:
            self._uid         = uid
            self._reconnected = True
        else:
            # generate new uid, reset all other ID counters
            # FIXME: this will screw up counters for *concurrent* sessions, 
            #        as the ID generation is managed in a process singleton.
            self._uid = ru.generate_id('rp.session',  mode=ru.ID_PRIVATE)
            ru.reset_id_counters(prefix='rp.session', reset_all_others=True)

        if not self._cfg.get('session_id'): self._cfg['session_id'] = self._uid 
        if not self._cfg.get('owner')     : self._cfg['owner']      = self._uid 
        if not self._cfg.get('logdir')    : self._cfg['logdir']     = '%s/%s' \
                                                     % (os.getcwd(), self._uid)
        self._logdir = self._cfg['logdir']
        self._prof   = self._get_profiler(name=self._cfg['owner'])
        self._rep    = self._get_reporter(name=self._cfg['owner'])
        self._log    = self._get_logger  (name=self._cfg['owner'],
                                          level=self._cfg.get('debug'))

        if _connect:

            # we need a dburl to connect to.
            if not dburl:
                dburl = os.environ.get("RADICAL_PILOT_DBURL")

            if not dburl:
                dburl = self._cfg.get('default_dburl')

            if not dburl:
                dburl = self._cfg.get('dburl')

            if not dburl:
                # we forgive missing dburl on reconnect, but not otherwise
                raise RuntimeError("no database URL (set RADICAL_PILOT_DBURL)")  


        self._dburl = ru.Url(dburl)
        self._cfg['dburl'] = str(self._dburl)

        # now we have config and uid - initialize base class (saga session)
        rs.Session.__init__(self, uid=self._uid)


        # ----------------------------------------------------------------------
        # create new session
        if _connect:
            self._log.info("using database %s" % self._dburl)

            # if the database url contains a path element, we interpret that as
            # database name (without the leading slash)
            if  not self._dburl.path         or \
                self._dburl.path[0]   != '/' or \
                len(self._dburl.path) <=  1  :
                if not uid:
                    # we fake reconnnect if no DB is available -- but otherwise we
                    # really really need a db connection...
                    raise ValueError("incomplete DBURL '%s' no db name!" % self._dburl)

        if not self._reconnected:
            self._prof.prof('session_start', uid=self._uid)
            self._rep.info ('<<new session: ')
            self._rep.plain('[%s]' % self._uid)
            self._rep.info ('<<database   : ')
            self._rep.plain('[%s]' % self._dburl)

        self._load_resource_configs()

        self._rec = os.environ.get('RADICAL_PILOT_RECORD_SESSION')
        if self._rec:
            # NOTE: Session recording cannot handle reconnected sessions, yet.
            #       We thus turn it off here with a warning
            if self._reconnected:
                self._log.warn("no session recording on reconnected session")

            else:
                # append session ID to recording path
                self._rec = "%s/%s" % (self._rec, self._uid)

                # create recording path and record session
                os.system('mkdir -p %s' % self._rec)
                ru.write_json({'dburl': str(self.dburl)}, 
                              "%s/session.json" % self._rec)
                self._log.info("recording session in %s" % self._rec)


        # create/connect database handle
        try:
            self._dbs = DBSession(sid=self.uid, dburl=str(self._dburl),
                                  cfg=self._cfg, logger=self._log, 
                                  connect=_connect)

            # from here on we should be able to close the session again
            self._log.info("New Session created: %s." % self.uid)

        except Exception, ex:
            self._rep.error(">>err\n")
            self._log.exception('session create failed')
            raise RuntimeError("Couldn't create new session (database URL '%s' incorrect?): %s" \
                            % (dburl, ex))  

        # the session must not carry bridge and component handles across forks
        ru.atfork(self._atfork_prepare, self._atfork_parent, self._atfork_child)

        # if bridges and components are specified in the config, start them
        ruc = rpu.Component
        self._bridges    = ruc.start_bridges   (self._cfg, self, self._log)
        self._components = ruc.start_components(self._cfg, self, self._log)
        self.is_valid()

        # at this point we have a DB connection, logger, etc, and can record
        # some metadata
        self._log.info('radical.pilot version: %s' % rp_version_detail)
        self._log.info('radical.saga  version: %s' % rs.version_detail)
        self._log.info('radical.utils version: %s' % ru.version_detail)

        py_version_detail = sys.version.replace("\n", " ")
        self.inject_metadata({'radical_stack' : {'rp': rp_version_detail,
                                                 'rs': rs.version_detail,
                                                 'ru': ru.version_detail,
                                                 'py': py_version_detail}})


        # FIXME: make sure the above code results in a usable session on
        #        reconnect
        self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def _atfork_prepare(self): 
        pass

    def _atfork_parent(self) :
        pass

    def _atfork_child(self)  : 
        self._components = list()
        self._bridges    = list()
        self._to_close   = list()
        self._to_stop    = list()
        self._to_destroy = list()

    
    # --------------------------------------------------------------------------
    # Allow Session to function as a context manager in a `with` clause
    def __enter__(self):
        return self


    # --------------------------------------------------------------------------
    # Allow Session to function as a context manager in a `with` clause
    def __exit__(self, type, value, traceback):

        # FIXME: use cleanup_on_close, terminate_on_close attributes
        self.close()


    # --------------------------------------------------------------------------
    #
    def is_valid(self, term=True):

        # don't check validity during termination
        if self._closed:
            return True

        # if we check any manager or agent, it will likely also check the
        # session in turn.  We break that loop here.
        self._valid_iter += 1

        try:
            if self._valid_iter >= 2:
                # we are too deep - abort this line or tests
                return True

            if self._valid:
                for _,umgr in self._umgrs.iteritems():
                    if not umgr.is_valid(term):
                        self._valid = False
                        break

            if self._valid:
                for _,pmgr in self._pmgrs.iteritems():
                    if not pmgr.is_valid(term):
                        self._valid = False
                        break

            if self._valid:
                for bridge in self._bridges:
                    if not bridge.is_valid(term):
                        self._valid = False
                        break

            if self._valid:
                for component in self._components:
                    if not component.is_valid(term):
                        self._valid = False
                        break

        finally:
            pass

        if not self._valid and term:
            self._log.warn("session %s is invalid" % self.uid)
            self.close()
          # raise RuntimeError("session %s is invalid" % self.uid)

        return self._valid


    # --------------------------------------------------------------------------
    #
    def _load_resource_configs(self):

        self.is_valid()

        self._prof.prof('config_parser_start', uid=self._uid)

        # Loading all "default" resource configurations
        module_path  = os.path.dirname(os.path.abspath(__file__))
        default_cfgs = "%s/configs/resource_*.json" % module_path
        config_files = glob.glob(default_cfgs)

        for config_file in config_files:

            try:
                self._log.info("Load resource configurations from %s" % config_file)
                rcs = ResourceConfig.from_file(config_file)
            except Exception as e:
                self._log.exception("skip config file %s: %s" % (config_file, e))
                raise RuntimeError('config error (%s) - abort' % e)

            for rc in rcs:
                self._log.info("Load resource configurations for %s" % rc)
                self._resource_configs[rc] = rcs[rc].as_dict() 
                self._log.debug('read rcfg for %s (%s)', 
                        rc, self._resource_configs[rc].get('cores_per_node'))

        home         = os.environ.get('HOME', '')
        user_cfgs    = "%s/.radical/pilot/configs/resource_*.json" % home
        config_files = glob.glob(user_cfgs)

        for config_file in config_files:

            try:
                rcs = ResourceConfig.from_file(config_file)
            except Exception as e:
                self._log.exception("skip config file %s: %s" % (config_file, e))
                raise RuntimeError('config error (%s) - abort' % e)

            for rc in rcs:
                self._log.info("Load resource configurations for %s" % rc)

                if rc in self._resource_configs:
                    # config exists -- merge user config into it
                    ru.dict_merge(self._resource_configs[rc],
                                  rcs[rc].as_dict(),
                                  policy='overwrite')
                else:
                    # new config -- add as is
                    self._resource_configs[rc] = rcs[rc].as_dict() 

                self._log.debug('fix  rcfg for %s (%s)', 
                        rc, self._resource_configs[rc].get('cores_per_node'))

        default_aliases = "%s/configs/resource_aliases.json" % module_path
        self._resource_aliases = ru.read_json_str(default_aliases)['aliases']

        # check if we have aliases to merge
        usr_aliases = '%s/.radical/pilot/configs/resource_aliases.json' % home
        if os.path.isfile(usr_aliases):
            ru.dict_merge(self._resource_aliases,
                          ru.read_json_str(usr_aliases).get('aliases', {}),
                          policy='overwrite')

        self._prof.prof('config_parser_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def close(self, cleanup=False, terminate=True, download=False):
        """Closes the session.

        All subsequent attempts access objects attached to the session will 
        result in an error. If cleanup is set to True (default) the session
        data is removed from the database.

        **Arguments:**
            * **cleanup** (`bool`): Remove session from MongoDB (implies * terminate)
            * **terminate** (`bool`): Shut down all pilots associated with the session. 

        **Raises:**
            * :class:`radical.pilot.IncorrectState` if the session is closed
              or doesn't exist. 
        """

        # close only once
        if self._closed:
            return

        self._rep.info('closing session %s' % self._uid)
        self._log.debug("session %s closing", self._uid)
        self._prof.prof("session_close", uid=self._uid)

        # set defaults
        if cleanup   == None: cleanup   = True
        if terminate == None: terminate = True

        if  cleanup:
            # cleanup implies terminate
            terminate = True

        for umgr_uid,umgr in self._umgrs.iteritems():
            self._log.debug("session %s closes umgr   %s", self._uid, umgr_uid)
            umgr.close()
            self._log.debug("session %s closed umgr   %s", self._uid, umgr_uid)

        for pmgr_uid,pmgr in self._pmgrs.iteritems():
            self._log.debug("session %s closes pmgr   %s", self._uid, pmgr_uid)
            pmgr.close(terminate=terminate)
            self._log.debug("session %s closed pmgr   %s", self._uid, pmgr_uid)

        for comp in self._components:
            self._log.debug("session %s closes comp   %s", self._uid, comp.uid)
            comp.stop()
            comp.join()
            self._log.debug("session %s closed comp   %s", self._uid, comp.uid)

        for bridge in self._bridges:
            self._log.debug("session %s closes bridge %s", self._uid, bridge.uid)
            bridge.stop()
            bridge.join()
            self._log.debug("session %s closed bridge %s", self._uid, bridge.uid)

        if self._dbs:
            self._log.debug("session %s closes db (%s)", self._uid, cleanup)
            self._dbs.close(delete=cleanup)

        self._log.debug("session %s closed (delete=%s)", self._uid, cleanup)
        self._prof.prof("session_stop", uid=self._uid)
        self._prof.close()

        # support GC
        for x in self._to_close: 
            try:    x.close()
            except: pass
        for x in self._to_stop:
            try:    x.stop()
            except: pass
        for x in self._to_destroy:
            try:    x.destroy()
            except: pass

        self._closed = True
        self._valid = False

        # after all is said and done, we attempt to download the pilot log- and
        # profiles, if so wanted
        if download:

            self._prof.prof("session_fetch_start", uid=self._uid)
            self._log.debug('start download')
            tgt = os.getcwd()
            self.fetch_json    (tgt='%s/%s' % (tgt, self.uid))
            self.fetch_profiles(tgt=tgt)
            self.fetch_logfiles(tgt=tgt)

            self._prof.prof("session_fetch_stop", uid=self._uid)

        self._rep.info('<<session lifetime: %.1fs' % (self.closed - self.created))
        self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object.
        """

        self.is_valid()

        object_dict = {
            "uid"       : self._uid,
            "created"   : self.created,
            "connected" : self.connected,
            "closed"    : self.closed,
            "dburl"     : str(self.dburl),
            "cfg"       : copy.deepcopy(self._cfg)
        }
        return object_dict


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the object.
        """
        return str(self.as_dict())


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def logdir(self):
        return self._logdir


    # --------------------------------------------------------------------------
    #
    @property
    def dburl(self):
        return self._dburl


    # --------------------------------------------------------------------------
    #
    def get_db(self):

        self.is_valid()

        if self._dbs: return self._dbs.get_db()
        else        : return None


    
    # --------------------------------------------------------------------------
    #
    @property
    def created(self):
        """Returns the UTC date and time the session was created.
        """
        if self._dbs: return self._dbs.created
        else        : return None


    # --------------------------------------------------------------------------
    #
    @property
    def connected(self):
        """Returns the most recent UTC date and time the session was
        reconnected to.
        """
        if self._dbs: return self._dbs.connected
        else        : return None


    # -------------------------------------------------------------------------
    #
    @property
    def is_connected(self):

        self.is_valid()

        return self._dbs.is_connected


    # --------------------------------------------------------------------------
    #
    @property
    def closed(self):
        """
        Returns the time of closing
        """
        if self._dbs: return self._dbs.closed
        else        : return None


    # --------------------------------------------------------------------------
    #
    def _get_logger(self, name, level=None):
        """
        This is a thin wrapper around `ru.Logger()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """
        return ru.Logger(name=name, ns='radical.pilot', targets=['.'], 
                         path=self._logdir, level=level)


    # --------------------------------------------------------------------------
    #
    def _get_reporter(self, name):
        """
        This is a thin wrapper around `ru.Reporter()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """

        return ru.Reporter(name=name, ns='radical.pilot', targets=['stdout'],
                           path=self._logdir)


    # --------------------------------------------------------------------------
    #
    def _get_profiler(self, name):
        """
        This is a thin wrapper around `ru.Profiler()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """

        prof = ru.Profiler(name=name, ns='radical.pilot', path=self._logdir)

        return prof


    # --------------------------------------------------------------------------
    #
    def _get_reporter(self, name):
        """
        This is a thin wrapper around `ru.Reporter()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """

        if not self._reporter:
            self._reporter = ru.Reporter(name=name, ns='radical.pilot',
                                         targets=['stdout'], path=self._logdir)
        return self._reporter


    # --------------------------------------------------------------------------
    #
    def _get_profiler(self, name):
        """
        This is a thin wrapper around `ru.Profiler()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """

        prof = ru.Profiler(name=name, ns='radical.pilot', path=self._logdir)

        return prof


    # --------------------------------------------------------------------------
    #
    def inject_metadata(self, metadata):
        """
        Insert (experiment) metadata into an active session
        RP stack version info always get added.
        """

        self.is_valid()

        if not isinstance(metadata, dict):
            raise Exception("Session metadata should be a dict!")

        if self._dbs and self._dbs._c:
            self._dbs._c.update({'type'  : 'session',
                                 "uid"   : self.uid},
                                {"$push" : {"metadata": metadata}})


    # --------------------------------------------------------------------------
    #
    def _register_pmgr(self, pmgr):

        self.is_valid()

        self._dbs.insert_pmgr(pmgr.as_dict())
        self._pmgrs[pmgr.uid] = pmgr


    # --------------------------------------------------------------------------
    #
    def list_pilot_managers(self):
        """
        Lists the unique identifiers of all :class:`radical.pilot.PilotManager` 
        instances associated with this session.

        **Returns:**
            * A list of :class:`radical.pilot.PilotManager` uids (`list` of `strings`).
        """

        self.is_valid()
        return self._pmgrs.keys()


    # --------------------------------------------------------------------------
    #
    def get_pilot_managers(self, pmgr_uids=None):
        """ 
        returns known PilotManager(s).

        **Arguments:**

            * **pmgr_uids** [`string`]: 
              unique identifier of the PilotManager we want

        **Returns:**
            * One or more [:class:`radical.pilot.PilotManager`] objects.
        """

        self.is_valid()

        return_scalar = False
        if not isinstance(pmgr_uids, list):
            pmgr_uids     = [pmgr_uids]
            return_scalar = True

        if pmgr_uids: pmgrs = [self._pmgrs[uid] for uid in pmgr_uids]
        else        : pmgrs =  self._pmgrs.values()

        if return_scalar: return pmgrs[0]
        else            : return pmgrs


    # --------------------------------------------------------------------------
    #
    def _register_umgr(self, umgr):

        self.is_valid()

        self._dbs.insert_umgr(umgr.as_dict())
        self._umgrs[umgr.uid] = umgr


    # --------------------------------------------------------------------------
    #
    def list_unit_managers(self):
        """
        Lists the unique identifiers of all :class:`radical.pilot.UnitManager` 
        instances associated with this session.

        **Returns:**
            * A list of :class:`radical.pilot.UnitManager` uids (`list` of `strings`).
        """

        self.is_valid()
        return self._umgrs.keys()


    # --------------------------------------------------------------------------
    #
    def get_unit_managers(self, umgr_uids=None):
        """ 
        returns known UnitManager(s).

        **Arguments:**

            * **umgr_uids** [`string`]: 
              unique identifier of the UnitManager we want

        **Returns:**
            * One or more [:class:`radical.pilot.UnitManager`] objects.
        """

        self.is_valid()

        return_scalar = False
        if not isinstance(umgr_uids, list):
            umgr_uids     = [umgr_uids]
            return_scalar = True

        if umgr_uids: umgrs = [self._umgrs[uid] for uid in umgr_uids]
        else        : umgrs =  self._umgrs.values()

        if return_scalar: return umgrs[0]
        else            : return umgrs


    # -------------------------------------------------------------------------
    #
    def list_resources(self):
        '''
        Returns a list of known resource labels which can be used in a pilot
        description.  Not that resource aliases won't be listed.
        '''

        return sorted(self._resource_configs.keys())


    # -------------------------------------------------------------------------
    #
    def add_resource_config(self, resource_config):
        """Adds a new :class:`radical.pilot.ResourceConfig` to the PilotManager's 
           dictionary of known resources, or accept a string which points to
           a configuration file.

           For example::

                  rc = radical.pilot.ResourceConfig(label="mycluster")
                  rc.job_manager_endpoint = "ssh+pbs://mycluster
                  rc.filesystem_endpoint  = "sftp://mycluster
                  rc.default_queue        = "private"
                  rc.bootstrapper         = "default_bootstrapper.sh"

                  pm = radical.pilot.PilotManager(session=s)
                  pm.add_resource_config(rc)

                  pd = radical.pilot.ComputePilotDescription()
                  pd.resource = "mycluster"
                  pd.cores    = 16
                  pd.runtime  = 5 # minutes

                  pilot = pm.submit_pilots(pd)
        """

        self.is_valid()

        if isinstance(resource_config, basestring):

            # let exceptions fall through
            rcs = ResourceConfig.from_file(resource_config)

            for rc in rcs:
                self._log.info("Loaded resource configurations for %s" % rc)
                self._resource_configs[rc] = rcs[rc].as_dict() 
                self._log.debug('add  rcfg for %s (%s)', 
                        rc, self._resource_configs[rc].get('cores_per_node'))

        else:
            self._resource_configs[resource_config.label] = resource_config.as_dict()
            self._log.debug('Add  rcfg for %s (%s)', 
                    resource_config.label, 
                    self._resource_configs[resource_config.label].get('cores_per_node'))

    # -------------------------------------------------------------------------
    #
    def get_resource_config(self, resource, schema=None):
        """
        Returns a dictionary of the requested resource config
        """

        self.is_valid()

        if  resource in self._resource_aliases:
            self._log.warning("using alias '%s' for deprecated resource key '%s'" \
                              % (self._resource_aliases[resource], resource))
            resource = self._resource_aliases[resource]

        if  resource not in self._resource_configs:
            raise RuntimeError("Resource '%s' is not known." % resource)

        resource_cfg = copy.deepcopy(self._resource_configs[resource])
        self._log.debug('get rcfg 1 for %s (%s)',  resource, resource_cfg.get('cores_per_node'))

        if  not schema:
            if 'schemas' in resource_cfg:
                schema = resource_cfg['schemas'][0]

        if  schema:
            if  schema not in resource_cfg:
                raise RuntimeError("schema %s unknown for resource %s" \
                                  % (schema, resource))

            for key in resource_cfg[schema]:
                # merge schema specific resource keys into the
                # resource config
                resource_cfg[key] = resource_cfg[schema][key]

        self._log.debug('get rcfg 2 for %s (%s)',  resource, resource_cfg.get('cores_per_node'))

        return resource_cfg


    # -------------------------------------------------------------------------
    #
    def fetch_profiles(self, tgt=None, fetch_client=False):

        return rpu.fetch_profiles(self._uid, dburl=self.dburl, tgt=tgt, 
                                  session=self)


    # -------------------------------------------------------------------------
    #
    def fetch_logfiles(self, tgt=None, fetch_client=False):

        return rpu.fetch_logfiles(self._uid, dburl=self.dburl, tgt=tgt, 
                                  session=self)


    # -------------------------------------------------------------------------
    #
    def fetch_json(self, tgt=None, fetch_client=False):

        return rpu.fetch_json(self._uid, dburl=self.dburl, tgt=tgt,
                              session=self)



    # -------------------------------------------------------------------------
    #
    def _get_client_sandbox(self):
        """
        For the session in the client application, this is os.getcwd().  For the
        session in any other component, specifically in pilot components, the
        client sandbox needs to be read from the session config (or pilot
        config).  The latter is not yet implemented, so the pilot can not yet
        interpret client sandboxes.  Since pilot-side stagting to and from the
        client sandbox is not yet supported anyway, this seems acceptable
        (FIXME).
        """

        return self._client_sandbox


    # -------------------------------------------------------------------------
    #
    def _get_resource_sandbox(self, pilot):
        """
        for a given pilot dict, determine the global RP sandbox, based on the
        pilot's 'resource' attribute.
        """

        self.is_valid()

        # FIXME: this should get 'resource, schema=None' as parameters

        resource = pilot['description'].get('resource')
        schema   = pilot['description'].get('access_schema')

        if not resource:
            raise ValueError('Cannot get pilot sandbox w/o resource target')

        # the global sandbox will be the same for all pilots on any resource, so
        # we cache it
        with self._cache_lock:

            if resource not in self._cache['resource_sandbox']:

                # cache miss -- determine sandbox and fill cache
                rcfg   = self.get_resource_config(resource, schema)
                fs_url = rs.Url(rcfg['filesystem_endpoint'])

                # Get the sandbox from either the pilot_desc or resource conf
                sandbox_raw = pilot['description'].get('sandbox')
                if not sandbox_raw:
                    sandbox_raw = rcfg.get('default_remote_workdir', "$PWD")
        
                # If the sandbox contains expandables, we need to resolve those remotely.
                # NOTE: Note that this will only work for (gsi)ssh or shell based access mechanisms
                if '$' not in sandbox_raw and '`' not in sandbox_raw:
                    # no need to expand further
                    sandbox_base = sandbox_raw

                else:
                    js_url = rs.Url(rcfg['job_manager_endpoint'])
        
                    if 'ssh' in js_url.schema.split('+'):
                        js_url.schema = 'ssh'
                    elif 'gsissh' in js_url.schema.split('+'):
                        js_url.schema = 'gsissh'
                    elif 'fork' in js_url.schema.split('+'):
                        js_url.schema = 'fork'
                    elif '+' not in js_url.schema:
                        # For local access to queueing systems use fork
                        js_url.schema = 'fork'
                    else:
                        raise Exception("unsupported access schema: %s" % js_url.schema)
        
                    self._log.debug("rsup.PTYShell('%s')", js_url)
                    shell = rsup.PTYShell(js_url, self)
        
                    ret, out, err = shell.run_sync(' echo "WORKDIR: %s"' % sandbox_raw)
                    if ret == 0 and 'WORKDIR:' in out:
                        sandbox_base = out.split(":")[1].strip()
                        self._log.debug("sandbox base %s: '%s'", js_url, sandbox_base)
                    else:
                        raise RuntimeError("Couldn't get remote working directory.")
        
                # at this point we have determined the remote 'pwd' - the global sandbox
                # is relative to it.
                fs_url.path = "%s/radical.pilot.sandbox" % sandbox_base
        
                # before returning, keep the URL string in cache
                self._cache['resource_sandbox'][resource] = fs_url

            return self._cache['resource_sandbox'][resource]


    # --------------------------------------------------------------------------
    #
    def _get_session_sandbox(self, pilot):

        self.is_valid()

        # FIXME: this should get 'resource, schema=None' as parameters

        resource = pilot['description'].get('resource')

        if not resource:
            raise ValueError('Cannot get session sandbox w/o resource target')

        with self._cache_lock:

            if resource not in self._cache['session_sandbox']:

                # cache miss
                resource_sandbox      = self._get_resource_sandbox(pilot)
                session_sandbox       = rs.Url(resource_sandbox)
                session_sandbox.path += '/%s' % self.uid

                self._cache['session_sandbox'][resource] = session_sandbox

            return self._cache['session_sandbox'][resource]


    # --------------------------------------------------------------------------
    #
    def _get_pilot_sandbox(self, pilot):

        self.is_valid()

        # FIXME: this should get 'pid, resource, schema=None' as parameters

        self.is_valid()

        pilot_sandbox = pilot.get('pilot_sandbox')
        if str(pilot_sandbox):
            return rs.Url(pilot_sandbox)

        pid = pilot['uid']
        with self._cache_lock:
            if  pid in self._cache['pilot_sandbox']:
                return self._cache['pilot_sandbox'][pid]

        # cache miss
        session_sandbox     = self._get_session_sandbox(pilot)
        pilot_sandbox       = rs.Url(session_sandbox)
        pilot_sandbox.path += '/%s/' % pilot['uid']

        with self._cache_lock:
            self._cache['pilot_sandbox'][pid] = pilot_sandbox

        return pilot_sandbox


    # --------------------------------------------------------------------------
    #
    def _get_unit_sandbox(self, unit, pilot):

        self.is_valid()

        # we don't cache unit sandboxes, they are just a string concat.
        pilot_sandbox = self._get_pilot_sandbox(pilot)
        return "%s/%s/" % (pilot_sandbox, unit['uid'])


    # --------------------------------------------------------------------------
    #
    def _get_jsurl(self, pilot):
        '''
        get job service endpoint and hop URL for the pilot's target resource.
        '''

        self.is_valid()

        resrc   = pilot['description']['resource']
        schema  = pilot['description']['access_schema']
        rcfg    = self.get_resource_config(resrc, schema)

        js_url  = rs.Url(rcfg.get('job_manager_endpoint'))
        js_hop  = rs.Url(rcfg.get('job_manager_hop', js_url))

        # make sure the js_hop url points to an interactive access
        # TODO: this is an unreliable heuristics - we should require the js_hop
        #       URL to be specified in the resource configs.
        if   '+gsissh' in js_hop.schema or \
             'gsissh+' in js_hop.schema    : js_hop.schema = 'gsissh'
        elif '+ssh'    in js_hop.schema or \
             'ssh+'    in js_hop.schema    : js_hop.schema = 'ssh'
        else                               : js_hop.schema = 'fork'

        return js_url, js_hop


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def autopilot(user, passwd):

        import github3
        import random

        labels = 'type:autopilot'
        titles = ['+++ Out of Cheese Error +++',
                  '+++ Redo From Start! +++',
                  '+++ Mr. Jelly! Mr. Jelly! +++',
                  '+++ Melon melon melon',
                  '+++ Wahhhhhhh! Mine! +++',
                  '+++ Divide By Cucumber Error +++',
                  '+++ Please Reinstall Universe And Reboot +++',
                  '+++ Whoops! Here comes the cheese! +++',
                  '+++ End of Cheese Error +++',
                  '+++ Can Not Find Drive Z: +++',
                  '+++ Unknown Application Error +++',
                  '+++ Please Reboot Universe +++',
                  '+++ Year Of The Sloth +++',
                  '+++ error of type 5307 has occured +++',
                  '+++ Eternal domain error +++',
                  '+++ Error at Address Number 6, Treacle Mine Road +++']

        def excuse():
            cmd_fetch  = "telnet bofh.jeffballard.us 666 2>&1 "
            cmd_filter = "grep 'Your excuse is:' | cut -f 2- -d :"
            out        = ru.sh_callout("%s | %s" % (cmd_fetch, cmd_filter),
                                       shell=True)[0]
            return out.strip()


        github = github3.login(user, passwd)
        repo   = github.repository("radical-cybertools", "radical.pilot")

        title = 'autopilot: %s' % titles[random.randint(0, len(titles)-1)]

        print '----------------------------------------------------'
        print 'autopilot'

        for issue in repo.issues(labels=labels, state='open'):
            if issue.title == title:
                reply = 'excuse: %s' % excuse()
                issue.create_comment(reply)
                print '  resolve: %s' % reply
                return

        # issue not found - create
        body  = 'problem: %s' % excuse()
        issue = repo.create_issue(title=title, body=body, labels=[labels],
                                  assignee=user)
        print '  issue  : %s' % title
        print '  problem: %s' % body
        print '----------------------------------------------------'


# -----------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import copy
import stat
import time
import Queue
import tempfile
import threading
import traceback
import subprocess

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc

from .base import AgentExecutingComponent


# ==============================================================================
#
class ABDS(AgentExecutingComponent):

    # The name is rong based on the abstraction, but for the moment I do not
    # have any other ideas

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__ (self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

        from .... import pilot as rp

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING, 
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)

        self._cancel_lock    = threading.RLock()
        self._cus_to_cancel  = list()
        self._cus_to_watch   = list()
        self._watch_queue    = Queue.Queue ()

        self._pilot_id = self._cfg['pilot_id']

        # run watcher thread
        self._terminate = threading.Event()
        self._watcher   = threading.Thread(target=self._watch, name="Watcher")
        self._watcher.daemon = True
        self._watcher.start ()

        # The AgentExecutingComponent needs the LaunchMethods to construct
        # commands.
        self._task_launcher = rp.agent.LM.create(
                name    = self._cfg['task_launch_method'],
                cfg     = self._cfg,
                session = self._session)

        self._mpi_launcher = rp.agent.LM.create(
                name    = self._cfg['mpi_launch_method'],
                cfg     = self._cfg,
                session = self._session)

        self._cu_environment = self._populate_cu_environment()

        self.gtod   = "%s/gtod" % self._pwd
        self.tmpdir = tempfile.gettempdir()

        # if we need to transplant any original env into the CU, we dig the
        # respective keys from the dump made by bootstrap_0.sh
        self._env_cu_export = dict()
        if self._cfg.get('export_to_cu'):
            with open('env.orig', 'r') as f:
                for line in f.readlines():
                    if '=' in line:
                        k,v = line.split('=', 1)
                        key = k.strip()
                        val = v.strip()
                        if key in self._cfg['export_to_cu']:
                            self._env_cu_export[key] = val


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        # terminate watcher thread
        self._terminate.set()
        if self._watcher:
            self._watcher.join()


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_units':

            self._log.info("cancel_units command (%s)" % arg)
            with self._cancel_lock:
                self._cus_to_cancel.extend(arg['uids'])

        return True


    # --------------------------------------------------------------------------
    #
    def _populate_cu_environment(self):
        """Derive the environment for the cu's from our own environment."""

        # Get the environment of the agent
        new_env = copy.deepcopy(os.environ)

        #
        # Mimic what virtualenv's "deactivate" would do
        #
        old_path = new_env.pop('_OLD_VIRTUAL_PATH', None)
        if old_path:
            new_env['PATH'] = old_path

        old_ppath = new_env.pop('_OLD_VIRTUAL_PYTHONPATH', None)
        if old_ppath:
            new_env['PYTHONPATH'] = old_ppath

        old_home = new_env.pop('_OLD_VIRTUAL_PYTHONHOME', None)
        if old_home:
            new_env['PYTHON_HOME'] = old_home

        old_ps = new_env.pop('_OLD_VIRTUAL_PS1', None)
        if old_ps:
            new_env['PS1'] = old_ps

        new_env.pop('VIRTUAL_ENV', None)

        # Remove the configured set of environment variables from the
        # environment that we pass to Popen.
        for e in new_env.keys():
            env_removables = list()
            if self._mpi_launcher : env_removables += self._mpi_launcher.env_removables
            if self._task_launcher: env_removables += self._task_launcher.env_removables
            for r in  env_removables:
                if e.startswith(r):
                    new_env.pop(e, None)

        return new_env


    # --------------------------------------------------------------------------
    #
    def work(self, units):

        if not isinstance(units, list):
            units = [units]

        self.advance(units, rps.AGENT_SCHEDULING, publish=True, push=False)

        for unit in units:
            self._handle_unit(unit)

        self.advance(units, rps.AGENT_EXECUTING_PENDING, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def _handle_unit(self, cu):

        try:
            if cu['description']['mpi']:
                launcher = self._mpi_launcher
            else :
                launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher (mpi=%s)" % cu['description']['mpi'])

            self._log.debug("Launching unit with %s (%s).", launcher.name, launcher.launch_command)

            assert(cu['slots']) # FIXME: no assert, but check
            self._prof.prof('exec', msg='unit launch', uid=cu['uid'])

            # Start a new subprocess to launch the unit
            self.spawn(launcher=launcher, cu=cu)

        except Exception as e:
            # append the startup error to the units stderr.  This is
            # not completely correct (as this text is not produced
            # by the unit), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running CU")
            cu['stderr'] += "\nPilot cannot start compute unit:\n%s\n%s" \
                            % (str(e), traceback.format_exc())

            # Free the Slots, Flee the Flots, Ree the Frots!
            if cu['slots']:
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

            self.advance(cu, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, cu):

        self._prof.prof('spawn', msg='unit spawn', uid=cu['uid'])

        # NOTE: see documentation of cu['sandbox'] semantics in the ComputeUnit
        #       class definition.
        sandbox = '%s/%s' % (self._pwd, cu['uid'])

        # make sure the sandbox exists
        rpu.rec_makedir(sandbox)

        # prep stdout/err so that we can append w/o checking for None
        cu['stdout'] = ''
        cu['stderr'] = ''
        cu['workdir']=sandbox

        launch_script_name = '%s/%s.sh' % (sandbox, cu['uid'])

        with open(launch_script_name, "w") as launch_script:
            launch_script.write('#!/bin/sh\n\n')

            # Create string for environment variable setting
            env_string  = ''
            if cu['description']['environment']:
                for key,val in cu['description']['environment'].iteritems():
                    env_string += 'export %s="%s"\n' % (key, val)
            env_string += 'export RP_SESSION_ID="%s"\n'   % self._cfg['session_id']
            env_string += 'export RP_PILOT_ID="%s"\n'     % self._cfg['pilot_id']
            env_string += 'export RP_AGENT_ID="%s"\n'     % self._cfg['agent_name']
            env_string += 'export RP_SPAWNER_ID="%s"\n'   % self.uid
            env_string += 'export RP_UNIT_ID="%s"\n'      % cu['uid']
            env_string += 'export RP_GTOD="%s"\n'         % self.gtod
            if 'RADICAL_PILOT_PROFILE' in os.environ:
                env_string += 'export RP_PROF="%s/%s.prof"\n' % (sandbox, cu['uid'])
            # also add any env vars requested for export by the resource config
            for k,v in self._env_cu_export.iteritems():
                env_string += "export %s=%s\n" % (k,v)

            env_string += '''
prof(){
    if test -z "$RP_PROF"
    then
        return
    fi
    event=$1
    now=$($RP_GTOD)
    echo "$now,$event,unit_script,MainThread,$RP_UNIT_ID,AGENT_EXECUTING," >> $RP_PROF
}
'''

            # also add any env vars requested in the unit description
            if cu['description']['environment']:
                for key,val in cu['description']['environment'].iteritems():
                    env_string += 'export %s=%s\n' % (key, val)

            launch_script.write('\n# Environment variables\n%s\n' % env_string)

            launch_script.write('prof cu_start\n')
            launch_script.write('\n# Change to unit sandbox\ncd %s\n' % sandbox)
            launch_script.write('prof cu_cd_done\n')

            # Before the Big Bang there was nothing
            if cu['description']['pre_exec']:
                pre_exec_string = ''
                if isinstance(cu['description']['pre_exec'], list):
                    for elem in cu['description']['pre_exec']:
                        pre_exec_string += "%s\n" % elem
                else:
                    pre_exec_string += "%s\n" % cu['description']['pre_exec']
                # Note: extra spaces below are for visual alignment
                launch_script.write("\n# Pre-exec commands\n")
                launch_script.write('prof cu_pre_start\n')
                launch_script.write(pre_exec_string)
                launch_script.write('prof cu_pre_stop\n')

            # YARN pre execution folder permission change
            launch_script.write('\n## Changing Working Directory permissions for YARN\n')
            launch_script.write('old_perm="`stat -c %a .`"\n')
            launch_script.write('chmod -R 777 .\n')

            # The actual command line, constructed per launch-method
            try:
                self._log.debug("Launch Script Name %s",launch_script_name)
                launch_command, hop_cmd = launcher.construct_command(cu, launch_script_name)
                self._log.debug("Launch Command %s from %s", launch_command, launcher.name)

                if hop_cmd : cmdline = hop_cmd
                else       : cmdline = launch_script_name

            except Exception as e:
                msg = "Error in spawner (%s)" % e
                self._log.exception(msg)
                raise RuntimeError(msg)

            launch_script.write("\n# The command to run\n")
            launch_script.write('prof cu_exec_start\n')
            launch_script.write("%s\n" % launch_command)
            launch_script.write("RETVAL=$?\n")
            launch_script.write("\ncat Ystdout\n")
            launch_script.write('prof cu_exec_stop\n')

            # After the universe dies the infrared death, there will be nothing
            if cu['description']['post_exec']:
                post_exec_string = ''
                if isinstance(cu['description']['post_exec'], list):
                    for elem in cu['description']['post_exec']:
                        post_exec_string += "%s\n" % elem
                else:
                    post_exec_string += "%s\n" % cu['description']['post_exec']
                launch_script.write("\n# Post-exec commands\n")
                launch_script.write('prof cu_post_start\n')
                launch_script.write('%s\n' % post_exec_string)
                launch_script.write('prof cu_post_stop\n')

            # YARN pre execution folder permission change
            launch_script.write('\n## Changing Working Directory permissions for YARN\n')
            launch_script.write('chmod $old_perm .\n')

            launch_script.write("\n# Exit the script with the return code from the command\n")
            launch_script.write("exit $RETVAL\n")

        # done writing to launch script, get it ready for execution.
        st = os.stat(launch_script_name)
        os.chmod(launch_script_name, st.st_mode | stat.S_IEXEC)
        self._prof.prof('control', msg='launch script constructed', uid=cu['uid'])

        # prepare stdout/stderr
        stdout_file = cu['description'].get('stdout') or 'STDOUT'
        stderr_file = cu['description'].get('stderr') or 'STDERR'

        cu['stdout_file'] = os.path.join(sandbox, stdout_file)
        cu['stderr_file'] = os.path.join(sandbox, stderr_file)

        _stdout_file_h = open(cu['stdout_file'], "w+")
        _stderr_file_h = open(cu['stderr_file'], "w+")
        self._prof.prof('control', msg='stdout and stderr files created', uid=cu['uid'])

        self._log.info("Launching unit %s via %s in %s", cu['uid'], cmdline, sandbox)

        self._prof.prof('spawn', msg='spawning passed to popen', uid=cu['uid'])

        cu['proc'] = subprocess.Popen(args               = cmdline,
                                      bufsize            = 0,
                                      executable         = None,
                                      stdin              = None,
                                      stdout             = _stdout_file_h,
                                      stderr             = _stderr_file_h,
                                      preexec_fn         = None,
                                      close_fds          = True,
                                      shell              = True,
                                      cwd                = sandbox,
                                      env                = self._cu_environment,
                                      universal_newlines = False,
                                      startupinfo        = None,
                                      creationflags      = 0)

        self._prof.prof('spawn', msg='spawning passed to popen', uid=cu['uid'])
        self._watch_queue.put(cu)


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        try:
            cuid = self.uid.replace('Component', 'Watcher')
            self._prof.prof('run', uid=self._pilot_id)

            while not self._terminate.is_set():

                cus = list()

                try:

                    # we don't want to only wait for one CU -- then we would
                    # pull CU state too frequently.  OTOH, we also don't want to
                    # learn about CUs until all slots are filled, because then
                    # we may not be able to catch finishing CUs in time -- so
                    # there is a fine balance here.  Balance means 100 (FIXME).
                  # self._prof.prof('pull')
                    MAX_QUEUE_BULKSIZE = 100
                    while len(cus) < MAX_QUEUE_BULKSIZE :
                        cus.append (self._watch_queue.get_nowait())

                except Queue.Empty:

                    # nothing found -- no problem, see if any CUs finished
                    pass

                # add all cus we found to the watchlist
                for cu in cus :

                    self._prof.prof('passed', msg="ExecWatcher picked up unit", uid=cu['uid'])
                    self._cus_to_watch.append (cu)

                # check on the known cus.
                action = self._check_running()

                if not action and not cus :
                    # nothing happened at all!  Zzz for a bit.
                    time.sleep(self._cfg['db_poll_sleeptime'])

        except Exception as e:
            self._log.exception("Error in ExecWorker watch loop (%s)" % e)
            # FIXME: this should signal the ExecWorker for shutdown...

        self._prof.prof('stop', uid=self._pilot_id)
        self._prof.flush()


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self):

        action = 0

        for cu in self._cus_to_watch:
            
            sandbox = '%s/%s' % (self._pwd, cu['uid'])

            #-------------------------------------------------------------------
            # This code snippet reads the YARN application report file and if
            # the application is RUNNING it update the state of the CU with the
            # right time stamp. In any other case it works as it was.
            logfile = '%s/%s' % (sandbox, '/YarnApplicationReport.log')
            if cu['state']==rps.AGENT_EXECUTING_PENDING \
                    and os.path.isfile(logfile):

                yarnreport = open(logfile,'r')
                report_contents = yarnreport.readlines()
                yarnreport.close()

                for report_line in report_contents:
                    if report_line.find('RUNNING') != -1:
                        self._log.debug(report_contents)
                        line = report_line.split(',')
                        timestamp = (int(line[3].split('=')[1])/1000)
                        action += 1
                        proc = cu['proc']
                        self._log.debug('Proc Print {0}'.format(proc))
                        del(cu['proc'])  # proc is not json serializable
                        self.advance(cu, rps.AGENT_EXECUTING, publish=True, 
                                     push=False,timestamp=timestamp)
                        cu['proc']    = proc

                        # FIXME: Ioannis, what is this supposed to do?
                        # I wanted to update the state of the cu but keep it in the watching
                        # queue. I am not sure it is needed anymore.
                        index = self._cus_to_watch.index(cu)
                        self._cus_to_watch[index]=cu

            else :
                # poll subprocess object
                exit_code = cu['proc'].poll()
                now       = time.time()

                if exit_code is None:
                    # Process is still running

                    if cu['uid'] in self._cus_to_cancel:

                        # FIXME: there is a race condition between the state poll
                        # above and the kill command below.  We probably should pull
                        # state after kill again?

                        # We got a request to cancel this cu
                        action += 1
                        cu['proc'].kill()
                        cu['proc'].wait() # make sure proc is collected

                        with self._cancel_lock:
                            self._cus_to_cancel.remove(cu['uid'])

                        self._prof.prof('final', msg="execution canceled", uid=cu['uid'])

                        self._cus_to_watch.remove(cu)

                        del(cu['proc'])  # proc is not json serializable
                        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)
                        self.advance(cu, rps.CANCELED, publish=True, push=False)

                else:
                    self._prof.prof('exec', msg='execution complete', uid=cu['uid'])


                    # make sure proc is collected
                    cu['proc'].wait()

                    # we have a valid return code -- unit is final
                    action += 1
                    self._log.info("Unit %s has return code %s.", cu['uid'], exit_code)

                    cu['exit_code'] = exit_code

                    # Free the Slots, Flee the Flots, Ree the Frots!
                    self._cus_to_watch.remove(cu)
                    del(cu['proc'])  # proc is not json serializable
                    self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

                    if exit_code != 0:
                        # The unit failed - fail after staging output
                        self._prof.prof('final', msg="execution failed", uid=cu['uid'])
                        cu['target_state'] = rps.FAILED

                    else:
                        # The unit finished cleanly, see if we need to deal with
                        # output data.  We always move to stageout, even if there are no
                        # directives -- at the very least, we'll upload stdout/stderr
                        self._prof.prof('final', msg="execution succeeded", uid=cu['uid'])
                        cu['target_state'] = rps.DONE

                    self.advance(cu, rps.AGENT_STAGING_OUTPUT_PENDING, publish=True, push=True)

        return action






import os
import sys
import copy
import time
import errno
import datetime
import pymongo
import netifaces
import threading
import multiprocessing

import radical.utils as ru
from   radical.pilot.states import *

# ------------------------------------------------------------------------------
#
# max number of cu out/err chars to push to tail
MAX_IO_LOGLENGTH = 1024
def tail(txt, maxlen=MAX_IO_LOGLENGTH):

    # shorten the given string to the last <n> characters, and prepend
    # a notification.  This is used to keep logging information in mongodb
    # manageable(the size of mongodb documents is limited).

    if not txt:
        return txt

    if len(txt) > maxlen:
        return "[... CONTENT SHORTENED ...]\n%s" % txt[-maxlen:]
    else:
        return txt


# ------------------------------------------------------------------------------
#
def get_rusage():

    import resource

    self_usage  = resource.getrusage(resource.RUSAGE_SELF)
    child_usage = resource.getrusage(resource.RUSAGE_CHILDREN)

    rtime = time.time()
    utime = self_usage.ru_utime  + child_usage.ru_utime
    stime = self_usage.ru_stime  + child_usage.ru_stime
    rss   = self_usage.ru_maxrss + child_usage.ru_maxrss

    return "real %3f sec | user %.3f sec | system %.3f sec | mem %.2f kB" \
         % (rtime, utime, stime, rss)


# ------------------------------------------------------------------------------
#
def rec_makedir(target):

    # recursive makedir which ignores errors if dir already exists

    try:
        os.makedirs(target)

    except OSError as e:
        # ignore failure on existing directory
        if e.errno == errno.EEXIST and os.path.isdir(os.path.dirname(target)):
            pass
        else:
            raise


# ------------------------------------------------------------------------------
#
_hostip = None
def hostip(req=None, black_list=None, pref_list=None, logger=None):
    """
    Look up the ip number for a given requested interface name.
    If interface is not given, do some magic.
    """

    # we only determine hostip once
    global _hostip
    if _hostip:
        return _hostip

    # List of interfaces that we probably dont want to bind to by default
    if not black_list:
        black_list = ['sit0', 'lo']

    # Known intefaces in preferred order
    if not pref_list:
        pref_list = [
            'ipogif0', # Cray's
            'br0'      # SuperMIC
        ]

    gateways = netifaces.gateways()
    if  not 'default' in gateways or \
        not gateways['default']:
        return '127.0.0.1'

    # we always add the currently used interface to the preferred ones
    default = gateways['default'][netifaces.AF_INET][1]
    if default not in pref_list:
        pref_list.append(default)

    # Get a list of all network interfaces
    all = netifaces.interfaces()

    if logger:
        logger.debug("Network interfaces detected: %s", all)

    pref = None
    # If we got a request, see if it is in the list that we detected
    if req and req in all:
        # Requested is available, set it
        pref = req
    else:
        # No requested or request not found, create preference list
        potentials = [iface for iface in all if iface not in black_list]

    # If we didn't select an interface already
    if not pref:
        # Go through the sorted list and see if it is available
        for iface in pref_list:
            if iface in all:
                # Found something, get out of here
                pref = iface
                break

    # If we still didn't find something, grab the first one from the
    # potentials if it has entries
    if not pref and potentials:
        pref = potentials[0]

    # If there were no potentials, see if we can find one in the blacklist
    if not pref:
        for iface in black_list:
            if iface in all:
                pref = iface

    if logger:
        logger.debug("Network interfaces selected: %s", pref)

    # Use IPv4, because, we can ...
    af = netifaces.AF_INET
    ip = netifaces.ifaddresses(pref)[af][0]['addr']

    # we fall back to localhost
    if not ip:
        ip = '127.0.0.1'

    if logger:
        logger.debug("Network ip address detected: %s", ip)

    # cache for next invocation
    _hostip = ip
    return ip


# ----------------------------------------------------------------------------------
#
def create_tar(tgt, dnames):
    '''
    Create a tarball on the file system which contains all given directories
    '''

    uid   = os.getuid()
    gid   = os.getgid()
    mode  = 16893
    mtime = time.time()

    fout  = open(tgt, 'wb')

    def rpad(s, size):
        return s + (size - len(s)) * '\0'
    
    def write_dir(path):
        data  = rpad(path, 100) \
              + rpad('%o' % mode,   8) \
              + rpad('%o' % uid,    8) \
              + rpad('%o' % gid,    8) \
              + rpad('%o' % 0,     12) \
              + rpad('%o' % mtime, 12) \
              + 8 * '\0' + '5'
        cksum = 256 + sum(ord(h) for h in data)
        data  = rpad(data  , 512)
        data  = data  [:-364] + '%06o\0' % cksum + data[-357:]
        fout.write(data)

    for dname in dnames:
        write_dir(dname)

    fout.close()


# ----------------------------------------------------------------------------------





import os
import sys

import radical.utils as ru

from .constants import *


# ------------------------------------------------------------------------------
#
def expand_description(descr):
    """
    convert any simple, string based staging directive in the description into 
    its dictionary equivalent

    In this context, the following kinds of expansions are performed:

      in:  ['input.dat'] 
      out: {'source' : 'client:///input.dat', 
            'target' : 'unit:///input.dat',
            'action' : rp.TRANSFER}

      in:  ['input.dat > staged.dat']
      out: {'source' : 'client:///input.dat', 
            'target' : 'unit:///staged.dat',
            'action' : rp.TRANSFER}

    This method changes the given description in place - repeated calls on the
    same description instance will have no effect.  However, we expect this
    method to be called only once during unit construction.
    """

    if descr.get('input_staging')  is None: descr['input_staging']  = list()
    if descr.get('output_staging') is None: descr['output_staging'] = list()

    descr['input_staging' ] = expand_staging_directives(descr['input_staging' ])
    descr['output_staging'] = expand_staging_directives(descr['output_staging'])


# ------------------------------------------------------------------------------
#
def expand_staging_directives(sds):
    """
    Take an abbreviated or compressed staging directive and expand it.
    """

    if not sds:
        return []

    if not isinstance(sds, list):
        sds = [sds]

    ret = list()
    for sd in sds:

        if isinstance(sd, basestring):
            # We detected a string, convert into dict.  The interpretation
            # differs depending of redirection characters being present in the
            # string.

            if   '>>' in sd: src, tgt = sd.split('>>', 2)
            elif '>'  in sd: src, tgt = sd.split('>' , 2)
            elif '<<' in sd: tgt, src = sd.split('<<', 2)
            elif '<'  in sd: tgt, src = sd.split('<' , 2)
            else           : src, tgt = sd, os.path.basename(ru.Url(sd).path)

            expanded = {'uid':      ru.generate_id('sd'),
                        'source':   src.strip(),
                        'target':   tgt.strip(),
                        'action':   DEFAULT_ACTION,
                        'flags':    DEFAULT_FLAGS,
                        'priority': DEFAULT_PRIORITY}

        elif isinstance(sd, dict):

            # sanity check on dict syntax
            valid_keys = ['source', 'target', 'action', 'flags', 'priority']
            for k in sd:
                if k not in valid_keys:
                    raise ValueError('"%s" is invalid on staging directive' % k)

            source   = sd.get('source')
            target   = sd.get('target',   os.path.basename(ru.Url(source).path))
            action   = sd.get('action',   DEFAULT_ACTION)
            flags    = sd.get('flags',    DEFAULT_FLAGS)
            priority = sd.get('priority', DEFAULT_PRIORITY)

            if not source:
                raise Exception("Staging directive dict has no source member!")

            # RCT flags should always be rendered as OR'ed integers - but old
            # versions of the RP API rendered them as list of strings.  We
            # convert to the integer version for backward compatibility - but we
            # complain loudly if we find actual strings.
            if isinstance(flags, list):
                int_flags = 0
                for flag in flags:
                    if isinstance(flags, basestring):
                        raise ValueError('"%s" is no valid RP constant' % flag)
                    int_flags != flag
                flags = int_flags

            elif isinstance(flags, basestring):
                raise ValueError('use RP constants for staging flags!')

            expanded = {'uid':      ru.generate_id('sd'),
                        'source':   source,
                        'target':   target,
                        'action':   action,
                        'flags':    flags,
                        'priority': priority}

        else:
            raise Exception("Unknown type of staging directive: %s (%s)" % (sd, type(sd)))

        # we warn the user when  src or tgt are using the deprecated
        # `staging://` schema
        if str(expanded['source']).startswith('staging://'):
            sys.stderr.write('staging:// schema is deprecated - use pilot://\n')
            expanded['source'] = str(expanded['source']).replace('staging://', 
                                                                 'pilot://')

        if str(expanded['target']).startswith('staging://'):
            sys.stderr.write('staging:// schema is deprecated - use pilot://\n')
            expanded['target'] = str(expanded['target']).replace('staging://', 
                                                                 'pilot://')

        ret.append(expanded)

    return ret


# ------------------------------------------------------------------------------
#
def complete_url(path, context, log=None):
    '''
    Some paths in data staging directives are to be interpreted relative to
    certain locations, namely relative to

        * `client://`  : the client's working directory
        * `resource://`: the RP    sandbox on the target resource
        * `pilot://`   : the pilot sandbox on the target resource
        * `unit://`    : the unit  sandbox on the target resource

    For the above schemas, we interpret `schema://` the same as `schema:///`,
    ie. we treat this as a namespace, not as location qualified by a hostname.

    The `context` parameter is expected to be a dict which provides a set of
    URLs to be used to expand the path.

    Other URL schemas are left alone, any other strings are interpreted as
    path in the context of `pwd`.

    The method returns an instance of ru.Url.  Note that URL parsing is not
    really cheap, so this method should be used conservatively.
    '''

    # FIXME: consider evaluation of env vars

    purl = ru.Url(path)

    log.debug('<- %s (%s)', path, type(path))
    log.debug('   %s', purl)

    str_path = str(path)

    # we always want a schema, and fall back to file:// or pwd://, depending if
    # the path is absolute or relative.  Note that pwd:// is interpreted in the
    # context of the staging component, which may live on the client or target
    # resource -- we make no attempt at expanding the path at this point.
    # We further assume that the user knows what she is doing when using
    # absolute paths, and make no attempts to verify those either.
    if not purl.schema:
        if str_path.startswith('/'):
            purl.schema = 'file'
        else:
            purl.schema = 'pwd'

    schema = purl.schema

    if schema == 'client': 
        # 'client' is 'pwd' in client context.  
        # We don't check context though.
        schema = 'pwd'  

    log.debug('   %s', schema)
    if schema in context.keys():

        # we interpret any hostname as part of the path element
        if   purl.host and purl.path: ppath = '%s/%s' % (purl.host, purl.path)
        elif purl.host              : ppath =    '%s' % (           purl.host)
        elif purl.path              : ppath =    '%s' % (           purl.path)
        else                        : ppath =     '.'

        if schema not in context:
            raise ValueError('cannot expand schema (%s) for staging' % schema)

        log.debug('   expand with %s', context[schema])
        ret = ru.Url(context[schema])

        if schema in ['resource', 'pilot']:
            # use a dedicated staging area dir
            ret.path += '/staging_area'

        ret.path += '/%s' % ppath
        purl      = ret

    # if not schema is set, assume file:// on localhost
    if not purl.schema:
        purl.schema = 'file'

    log.debug('-> %s', purl)
    return purl


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from .base import LaunchMethod


# ==============================================================================
#
# ccmrun: Cluster Compatibility Mode job launcher for Cray systems
#
class MPIRunCCMRun(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        LaunchMethod.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self.launch_command = ru.which([
            'mpirun',            # General case
            'mpirun_rsh',        # Gordon @ SDSC
            'mpirun-mpich-mp',   # Mac OSX MacPorts
            'mpirun-openmpi-mp'  # Mac OSX MacPorts
        ])

        self.ccmrun_command = ru.which([
            'ccmrun',            # General case
        ])

        if not self.ccmrun_command:
            raise RuntimeError("ccmrun not found!")

        self.mpi_version, self.mpi_flavor = self._get_mpi_info(self.launch_command)


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_env     = cud.get('environment') or dict()
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        # Construct the executable and arguments
        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
        env_list   = self.EXPORT_ENV_VARIABLES + task_env.keys()
        if env_list:

            if self.mpi_flavor == self.MPI_FLAVOR_HYDRA:
                env_string = '-envlist "%s"' % ','.join(env_list)

            elif self.mpi_flavor == self.MPI_FLAVOR_OMPI:
                for var in env_list:
                    env_string += '-x "%s" ' % var

        if 'nodes' not in slots:
            raise RuntimeError('insufficient information to launch via %s: %s'
                              % (self.name, slots))

        # Extract all the hosts from the slots
        # TODO: is there any use in using $HOME/.crayccm/ccm_nodelist.$JOBID?
        hostlist = list()
        for node in slots['nodes']:
            for cpu_proc in node['core_map']:
                hostlist.append(node['name'])
            for gpu_proc in node['gpu_map']:
                hostlist.append(node['name'])
        hosts_string = ",".join(hostlist)

        command = "%s %s -np %d -host %s %s %s" % \
                  (self.ccmrun_command, self.launch_command, len(hostlist), 
                   hosts_string, env_string, task_command)

        return command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import hostlist

from base import LRMS


# ==============================================================================
#
class Slurm(LRMS):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        LRMS.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        slurm_nodelist = os.environ.get('SLURM_NODELIST')
        if slurm_nodelist is None:
            msg = "$SLURM_NODELIST not set!"
            self._log.error(msg)
            raise RuntimeError(msg)

        # Parse SLURM nodefile environment variable
        slurm_nodes = hostlist.expand_hostlist(slurm_nodelist)
        self._log.info("Found SLURM_NODELIST %s. Expanded to: %s", slurm_nodelist, slurm_nodes)

        # $SLURM_NPROCS = Total number of cores allocated for the current job
        slurm_nprocs_str = os.environ.get('SLURM_NPROCS')
        if slurm_nprocs_str is None:
            msg = "$SLURM_NPROCS not set!"
            self._log.error(msg)
            raise RuntimeError(msg)
        else:
            slurm_nprocs = int(slurm_nprocs_str)

        # $SLURM_NNODES = Total number of (partial) nodes in the job's resource allocation
        slurm_nnodes_str = os.environ.get('SLURM_NNODES')
        if slurm_nnodes_str is None:
            msg = "$SLURM_NNODES not set!"
            self._log.error(msg)
            raise RuntimeError(msg)
        else:
            slurm_nnodes = int(slurm_nnodes_str)

        # $SLURM_CPUS_ON_NODE = Number of cores per node (physically)
        slurm_cpus_on_node_str = os.environ.get('SLURM_CPUS_ON_NODE')
        if slurm_cpus_on_node_str is None:
            msg = "$SLURM_CPUS_ON_NODE not set!"
            self._log.error(msg)
            raise RuntimeError(msg)
        else:
            slurm_cpus_on_node = int(slurm_cpus_on_node_str)

        # Verify that $SLURM_NPROCS <= $SLURM_NNODES * $SLURM_CPUS_ON_NODE
        if not slurm_nprocs <= slurm_nnodes * slurm_cpus_on_node:
            self._log.warning("$SLURM_NPROCS(%d) <= $SLURM_NNODES(%d) * $SLURM_CPUS_ON_NODE(%d)",
                            slurm_nprocs, slurm_nnodes, slurm_cpus_on_node)

        # Verify that $SLURM_NNODES == len($SLURM_NODELIST)
        if slurm_nnodes != len(slurm_nodes):
            self._log.error("$SLURM_NNODES(%d) != len($SLURM_NODELIST)(%d)",
                           slurm_nnodes, len(slurm_nodes))

        # Report the physical number of cores or the total number of cores
        # in case of a single partial node allocation.
        self.cores_per_node = self._cfg.get('cores_per_node', 0)
        self.gpus_per_node  = self._cfg.get('gpus_per_node',  0)  # FIXME GPU
        self.lfs_per_node   = {'path' : self._cfg.get('lfs_path_per_node', None),
                               'size' : self._cfg.get('lfs_size_per_node', 0)
                              }

        if not self.cores_per_node:
            self.cores_per_node = min(slurm_cpus_on_node, slurm_nprocs)


        # node names are unique, so can serve as node uids
        self.node_list = [[node, node] for node in slurm_nodes]

        self.lm_info['cores_per_node'] = self.cores_per_node


# ------------------------------------------------------------------------------





import os
import sys
import zmq
import copy
import math
import time
import errno
import pprint
import msgpack

import Queue           as pyq
import setproctitle    as spt
import multiprocessing as mp

import radical.utils   as ru

from .misc import hostip as rpu_hostip


# --------------------------------------------------------------------------
# defines for pubsub roles
#
PUBSUB_PUB    = 'pub'
PUBSUB_SUB    = 'sub'
PUBSUB_BRIDGE = 'bridge'
PUBSUB_ROLES  = [PUBSUB_PUB, PUBSUB_SUB, PUBSUB_BRIDGE]

_USE_MULTIPART   = False  # send [topic, data] as multipart message
_BRIDGE_TIMEOUT  =     5  # how long to wait for bridge startup
_LINGER_TIMEOUT  =   250  # ms to linger after close
_HIGH_WATER_MARK =     0  # number of messages to buffer before dropping


# --------------------------------------------------------------------------
#
# zmq will (rightly) barf at interrupted system calls.  We are able to rerun
# those calls.
#
# FIXME: how does that behave wrt. tomeouts?  We probably should include
#        an explicit timeout parameter.
#
# kudos: https://gist.github.com/minrk/5258909
#
def _uninterruptible(f, *args, **kwargs):
    cnt = 0
    while True:
        cnt += 1
        try:
            return f(*args, **kwargs)
        except zmq.ContextTerminated as e:
            return None
        except zmq.ZMQError as e:
            if e.errno == errno.EINTR:
                if cnt > 10:
                    raise
                # interrupted, try again
                continue
            else:
                # real error, raise it
                raise


# ==============================================================================
#
# Notifications between components are based on pubsub channels.  Those channels
# have different scope (bound to the channel name).  Only one specific topic is
# predefined: 'state' will be used for unit state updates.
#
class Pubsub(ru.Process):

    def __init__(self, session, channel, role, cfg, addr=None):
        """
        Addresses are of the form 'tcp://host:port'.  Both 'host' and 'port' can
        be wildcards for BRIDGE roles -- the bridge will report the in and out
        addresses as obj.addr_in and obj.addr_out.
        """

        self._session = session
        self._channel = channel
        self._role    = role
        self._cfg     = copy.deepcopy(cfg)
        self._addr    = addr

        assert(self._role in PUBSUB_ROLES), 'invalid role %s' % self._role

        self._uid = "%s.%s" % (self._channel.replace('_', '.'), self._role)
        self._uid = ru.generate_id(self._uid)
        self._log = self._session._get_logger(name=self._uid, 
                         level=self._cfg.get('log_level'))

        # avoid superfluous logging calls in critical code sections
        if self._log.getEffectiveLevel() == 10: # logging.DEBUG:
            self._debug = True
        else:
            self._debug = False

        self._addr_in   = None  # bridge input  addr
        self._addr_out  = None  # bridge output addr

        self._q    = None
        self._in   = None
        self._out  = None
        self._ctx  = None

        if not self._addr:
            self._addr = 'tcp://*:*'

        self._log.info("create %s - %s - %s", self._channel, self._role, self._addr)

        super(Pubsub, self).__init__(name=self._uid, log=self._log)


        # ----------------------------------------------------------------------
        # behavior depends on the role...
        if self._role == PUBSUB_PUB:

            self._ctx = zmq.Context()
            self._session._to_destroy.append(self._ctx)

            self._q   = self._ctx.socket(zmq.PUB)
            self._q.linger = _LINGER_TIMEOUT
            self._q.hwm    = _HIGH_WATER_MARK
            self._q.connect(self._addr)
            self.start(spawn=False)


        # ----------------------------------------------------------------------
        elif self._role == PUBSUB_BRIDGE:

            # we expect bridges to always use a port wildcard. Make sure
            # that's the case
            elems = self._addr.split(':')
            if len(elems) > 2 and elems[2] and elems[2] != '*':
                raise RuntimeError('wildcard port (*) required for bridge addresses (%s)' \
                                % self._addr)

            self._pqueue = mp.Queue()
            self.start()

            try:
                [addr_in, addr_out] = self._pqueue.get(True, _BRIDGE_TIMEOUT)

                # store addresses
                self._addr_in  = ru.Url(addr_in)
                self._addr_out = ru.Url(addr_out)

                # use the local hostip for bridge addresses
                self._addr_in.host  = rpu_hostip()
                self._addr_out.host = rpu_hostip()

            except pyq.Empty as e:
                raise RuntimeError ("bridge did not come up! (%s)" % e)


        # ----------------------------------------------------------------------
        elif self._role == PUBSUB_SUB:

            self._ctx = zmq.Context()
            self._session._to_destroy.append(self._ctx)

            self._q   = self._ctx.socket(zmq.SUB)
            self._q.linger = _LINGER_TIMEOUT
            self._q.hwm    = _HIGH_WATER_MARK
            self._q.connect(self._addr)
            self.start(spawn=False)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel

    @property
    def role(self):
        return self._role

    @property
    def addr(self):
        return self._addr

    @property
    def addr_in(self):
        assert(self._role == PUBSUB_BRIDGE), 'addr_in only set on bridges'
        return self._addr_in

    @property
    def addr_out(self):
        assert(self._role == PUBSUB_BRIDGE), 'addr_out only set on bridges'
        return self._addr_out


    # --------------------------------------------------------------------------
    # 
    def ru_initialize_child(self):

        assert(self._role == PUBSUB_BRIDGE), 'only bridges can be started'

        self._uid = self._uid + '.child'
        self._log = self._session._get_logger(name=self._uid, 
                         level=self._cfg.get('log_level'))

        spt.setproctitle('rp.%s' % self._uid)
        self._log.info('start bridge %s on %s', self._uid, self._addr)

        self._ctx = zmq.Context()
        self._session._to_destroy.append(self._ctx)

        self._in  = self._ctx.socket(zmq.XSUB)
        self._in.linger = _LINGER_TIMEOUT
        self._in.hwm    = _HIGH_WATER_MARK
        self._in.bind(self._addr)

        self._out = self._ctx.socket(zmq.XPUB)
        self._out.linger = _LINGER_TIMEOUT
        self._out.hwm    = _HIGH_WATER_MARK
        self._out.bind(self._addr)

        # communicate the bridge ports to the parent process
        _addr_in  = self._in.getsockopt( zmq.LAST_ENDPOINT)
        _addr_out = self._out.getsockopt(zmq.LAST_ENDPOINT)

        self._pqueue.put([_addr_in, _addr_out])

        self._log.info('bound bridge %s to %s : %s', self._uid, _addr_in, _addr_out)

        # start polling for messages
        self._poll = zmq.Poller()
        self._poll.register(self._in,  zmq.POLLIN)
        self._poll.register(self._out, zmq.POLLIN)


    # --------------------------------------------------------------------------
    # 
    def ru_finalize_common(self):

        if self._q   : self._q  .close()
        if self._in  : self._in .close()
        if self._out : self._out.close()
        if self._ctx : self._ctx.destroy()


    # --------------------------------------------------------------------------
    # 
    def work_cb(self):

        _socks = dict(_uninterruptible(self._poll.poll, timeout=1000)) # timeout in ms

        if self._in in _socks:
            
            # if any incoming socket signals a message, get the
            # message on the subscriber channel, and forward it
            # to the publishing channel, no questions asked.
            if _USE_MULTIPART:
                msg = _uninterruptible(self._in.recv_multipart, flags=zmq.NOBLOCK)
                _uninterruptible(self._out.send_multipart, msg)
            else:
                msg = _uninterruptible(self._in.recv, flags=zmq.NOBLOCK)
                _uninterruptible(self._out.send, msg)
          # if self._debug:
          #     self._log.debug("-> %s", pprint.pformat(msg))


        if self._out in _socks:
            # if any outgoing socket signals a message, it's
            # likely a topic subscription.  We forward that on
            # the incoming channels to subscribe for the
            # respective messages.
            if _USE_MULTIPART:
                msg = _uninterruptible(self._out.recv_multipart)
                _uninterruptible(self._in.send_multipart, msg)
            else:
                msg = _uninterruptible(self._out.recv)
                _uninterruptible(self._in.send, msg)
          # if self._debug:
          #     self._log.debug("<- %s", pprint.pformat(msg))

        return True


    # --------------------------------------------------------------------------
    #
    def subscribe(self, topic):

        assert(self._role == PUBSUB_SUB), 'incorrect role on subscribe'

        topic = topic.replace(' ', '_')

      # self._log.debug("~~ %s", topic)
        _uninterruptible(self._q.setsockopt, zmq.SUBSCRIBE, topic)


    # --------------------------------------------------------------------------
    #
    def put(self, topic, msg):

        assert(self._role == PUBSUB_PUB), 'incorrect role on put'
        assert(isinstance(msg,dict)),     'invalide message type'

      # self._log.debug("?> %s", pprint.pformat(msg)

        topic = topic.replace(' ', '_')
        data  = msgpack.packb(msg) 

        if _USE_MULTIPART:
          # if self._debug:
          #     self._log.debug("-> %s", ([topic, pprint.pformat(msg)]))
            _uninterruptible(self._q.send_multipart, [topic, data])

        else:
          # if self._debug:
          #     self._log.debug("-> %s %s", topic, pprint.pformat(msg))
            _uninterruptible(self._q.send, "%s %s" % (topic, data))


    # --------------------------------------------------------------------------
    #
    def get(self):

        assert(self._role == PUBSUB_SUB), 'invalid role on get'

        # FIXME: add timeout to allow for graceful termination

        if _USE_MULTIPART:
            topic, data = _uninterruptible(self._q.recv_multipart)

        else:
            raw = _uninterruptible(self._q.recv)
            topic, data = raw.split(' ', 1)

        msg = msgpack.unpackb(data) 
      # if self._debug:
      #     self._log.debug("<- %s", ([topic, pprint.pformat(msg)]))
        return [topic, msg]


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout=None): # timeout in ms

        assert(self._role == PUBSUB_SUB), 'invalid role on get_nowait'

        if _uninterruptible(self._q.poll, flags=zmq.POLLIN, timeout=timeout):

            if _USE_MULTIPART:
                topic, data = _uninterruptible(self._q.recv_multipart, 
                                               flags=zmq.NOBLOCK)

            else:
                raw = _uninterruptible(self._q.recv)
                topic, data = raw.split(' ', 1)

            msg = msgpack.unpackb(data) 
          # if self._debug:
          #     self._log.debug("<< %s", ([topic, pprint.pformat(msg)]))
            return [topic, msg]

        else:
            return [None, None]


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys
import time
import signal

os.environ['RADICAL_PILOT_VERBOSE'] = 'REPORT'

import radical.pilot as rp
import radical.utils as ru

# ------------------------------------------------------------------------------
#
# READ the RADICAL-Pilot documentation: http://radicalpilot.readthedocs.org/
#
# ------------------------------------------------------------------------------

n_pilots =    100
n_units  =   1000
n_done   =      0

#------------------------------------------------------------------------------
#
if __name__ == '__main__':

    start = time.time()

    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if len(sys.argv) >= 2: resources = sys.argv[1:]
    else                 : resources = ['local.localhost']

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/examples/config.json' % os.path.dirname(os.path.abspath(__file__)))
        report.ok('>>ok\n')

        report.header('submit pilots')

        # Add a Pilot Manager. Pilot managers manage one or more ComputePilots.
        pmgr = rp.PilotManager(session=session)

        scheduler = rp.umgr.scheduler.SCHEDULER_BACKFILLING
      # scheduler = rp.umgr.scheduler.SCHEDULER_ROUND_ROBIN
      # scheduler = None

        # Register the ComputePilot in a UnitManager object.
        umgr = rp.UnitManager(session=session, scheduler=scheduler)
        def unit_cb(unit, state):
            if state in rp.FINAL:
                global n_done
                n_done += 1
                print '%5s: %4d - %4d - %5.1f - %s' % (state[0:5], n_done, n_units,
                        time.time()-start, unit.pilot)
          # if state in [rp.FAILED]:
          #     session.close()
        umgr.register_callback(unit_cb)

        pdescs = list()
        for resource in resources:

            # Define an [n]-core local pilot that runs for [x] minutes
            # Here we use a dict to initialize the description object
            for i in range(n_pilots):
              
               ch = None
               if 'osg' in resource:
                   ch = [
                      #  '(HAS_MODULES =?= TRUE)',
                      #  '~(HAS_CVMFS_oasis_opensciencegrid_org =?= TRUE)',
                         '!(FIU_HPCOSG_CE)',  # no network
                         '!(CIT_CMS_T2)',     # no network
                        ]
               pd_init = {
                       'resource'        : resource,
                       'cores'           :   1,   # pilot size
                       'runtime'         : 300,   # pilot runtime (min)
                       'exit_on_error'   : False,
                       'project'         : config[resource]['project'],
                       'queue'           : config[resource]['queue'],
                       'access_schema'   : config[resource]['schema'],
                       'cleanup'         : False,
                       'candidate_hosts' : ch
                       }
               pdesc = rp.ComputePilotDescription(pd_init)
               pdescs.append(pdesc)
       
        # Launch the pilot.
        pilots = pmgr.submit_pilots(pdescs)
        umgr.add_pilots(pilots)

        def pilot_cb(pilot, state):
            print 'pilot: %s - %s - %5.1f' % (pilot.uid, state, time.time()-start)
          # if state in rp.FINAL:
          #     sys.exit()
        for pilot in pilots:
            pilot.register_callback(pilot_cb)
       
      # pmgr.wait_pilots(state=rp.ACTIVE)
      # sys.exit()


        # Create a workload of ComputeUnits.
        # Each compute unit runs '/bin/date'.
        report.header('submit units')

        report.info('create %d unit description(s)\n\t' % n_units)

        cuds  = list()
        start = time.time()
        for i in range(n_units):

            # create a new CU description, and fill it.
            # Here we don't use dict initialization.
            cud = rp.ComputeUnitDescription()
            # trigger an error now and then
          # if i % 2: 
            if False:
                cud.executable = 'sleep'
                cud.arguments  = ['30']
            else:
                cud.executable = '/bin/echo'
                cud.arguments  = ['$RP_PILOT_ID']

            cuds.append(cud)
            report.progress()
        report.ok('>>ok\n')

        # Submit the previously created ComputeUnit descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning ComputeUnits to the ComputePilots.
        start  = time.time()
        units  = umgr.submit_units(cuds)
        stop   = time.time()
        print ' === > %s' % (stop-start)

        # Wait for all compute units to reach a final state (DONE, CANCELED or FAILED).
        report.header('gather results')
        umgr.wait_units()
    
        report.info('\n')
        for unit in units:
            report.plain('  * %s: %s, %5s' \
                        % (unit.uid, unit.state[:4], unit.pilot))
            if unit.state in [rp.DONE]:
                report.ok('>>ok\n')
            elif unit.state in [rp.FAILED]:
                report.error('>>err\n')
            else:
                report.warn('>>nok\n')


      #     if unit.state in [rp.FAILED, rp.CANCELED]:
      #         report.plain('  * %s: %s, exit: %5s, err: %35s' \
      #                 % (unit.uid, unit.state[:4], 
      #                    unit.exit_code, unit.stderr.strip()))
      #         report.error('>>err\n')
      #     else:
      #         report.plain('  * %s: %s, exit: %5s, out: %35s' \
      #                 % (unit.uid, unit.state[:4], 
      #                     unit.exit_code, unit.stdout.strip()))
      #         report.ok('>>ok\n')
    

    except Exception as e:
        # Something unexpected happened in the pilot code above
        session._log.exception('oops')
        report.error('caught Exception: %s\n' % e)
        raise
    
    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        report.warn('exit requested\n')
    
    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        report.header()
        if session:
            session.close(cleanup=False)

    report.header()


#-------------------------------------------------------------------------------





# ------------------------------------------------------------------------------
# we *first* import radical.utils, so that the monkeypatching of the logger has
# a chance to kick in before the logging module is pulled by any other 3rd party
# module, and also to monkeypatch `os.fork()` for the `atfork` functionality.
# we also get the version string at this point.
#
import os
import radical.utils as _ru

_root = "%s/.." % os.path.dirname(__file__)

version_short, version_detail, version_base, version_branch, \
        sdist_name, sdist_path = _ru.get_version(_root)
version = version_short


# ------------------------------------------------------------------------------
#
from .db_utils     import *
from .prof_utils   import *
from .misc         import *
from .queue        import *
from .pubsub       import *
from .session      import *
from .component    import *
from .slot_utils   import *


# ------------------------------------------------------------------------------

